@@ -30,7 +30,6 @@
 #include "gdk.h"		/* For gdk_rectangle_union() */
 #include "gdkpixmap.h"
 #include "gdkdrawable.h"
-#include "gdkpixmap.h"
 #include "gdkscreen.h"
 #include "gdkalias.h"
 
@@ -676,7 +675,9 @@ gdk_window_peek_children (GdkWindow *window)
  * easy to break GDK and/or GTK+, so you have to know what you're
  * doing. Pass %NULL for @window to get all events for all windows,
  * instead of events for a specific window.
- * 
+ *
+ * See gdk_display_add_client_message_filter() if you are interested
+ * in X ClientMessage events.
  **/
 void          
 gdk_window_add_filter (GdkWindow     *window,
@@ -981,10 +982,13 @@ gdk_window_begin_paint_region (GdkWindow *window,
   if (GDK_WINDOW_DESTROYED (window))
     return;
 
-  if (GDK_IS_PAINTABLE (private->impl) &&
-      GDK_PAINTABLE_GET_IFACE (private->impl)->begin_paint_region) 
+  if (GDK_IS_PAINTABLE (private->impl)) 
     {
-      GDK_PAINTABLE_GET_IFACE (private->impl)->begin_paint_region (GDK_PAINTABLE (private->impl), region);
+      GdkPaintableIface *iface = GDK_PAINTABLE_GET_IFACE (private->impl);
+
+      if (iface->begin_paint_region)
+        iface->begin_paint_region ((GdkPaintable*)private->impl, region);
+
       return;
     }
 
@@ -1049,10 +1053,12 @@ gdk_window_end_paint (GdkWindow *window)
   if (GDK_WINDOW_DESTROYED (window))
     return;
 
-  if (GDK_IS_PAINTABLE (private->impl) &&
-      GDK_PAINTABLE_GET_IFACE (private->impl)->end_paint) 
+  if (GDK_IS_PAINTABLE (private->impl))
     {
-      GDK_PAINTABLE_GET_IFACE (private->impl)->end_paint (GDK_PAINTABLE (private->impl));
+      GdkPaintableIface *iface = GDK_PAINTABLE_GET_IFACE (private->impl);
+
+      if (iface->end_paint)
+        iface->end_paint ((GdkPaintable*)private->impl);
       return;
     }
 
@@ -2126,8 +2132,7 @@ gdk_window_draw_trapezoids (GdkDrawable   *drawable,
   else
     gdk_draw_trapezoids (private->impl, gc, trapezoids, n_trapezoids);
   
-  if (new_trapezoids)
-    g_free (new_trapezoids);
+  g_free (new_trapezoids);
 
   RESTORE_GC (gc);
 }
@@ -2250,9 +2255,7 @@ static gboolean debug_updates = FALSE;
 static gboolean
 gdk_window_update_idle (gpointer data)
 {
-  GDK_THREADS_ENTER ();
   gdk_window_process_all_updates ();
-  GDK_THREADS_LEAVE ();
   
   return FALSE;
 }
@@ -2265,7 +2268,7 @@ gdk_window_schedule_update (GdkWindow *window)
 
   if (!update_idle)
     {
-      update_idle = g_idle_add_full (GDK_PRIORITY_REDRAW,
+      update_idle = gdk_threads_add_idle_full (GDK_PRIORITY_REDRAW,
 				     gdk_window_update_idle, NULL, NULL);
     }
 }
@@ -2428,10 +2431,13 @@ gdk_window_process_updates (GdkWindow *window,
   g_return_if_fail (window != NULL);
   g_return_if_fail (GDK_IS_WINDOW (window));
   
-  if (GDK_IS_PAINTABLE (private->impl) &&
-      GDK_PAINTABLE_GET_IFACE (private->impl)->process_updates)
+  if (GDK_IS_PAINTABLE (private->impl))
     {
-      GDK_PAINTABLE_GET_IFACE (private->impl)->process_updates (GDK_PAINTABLE (private->impl), update_children);
+      GdkPaintableIface *iface = GDK_PAINTABLE_GET_IFACE (private->impl);
+
+      if (iface->process_updates)
+        iface->process_updates ((GdkPaintable*)private->impl, update_children);
+
       return;
     }
   
@@ -2564,11 +2570,13 @@ gdk_window_invalidate_maybe_recurse (GdkWindow *window,
   if (private->input_only || !GDK_WINDOW_IS_MAPPED (window))
     return;
 
-  if (GDK_IS_PAINTABLE (private->impl) &&
-      GDK_PAINTABLE_GET_IFACE (private->impl)->invalidate_maybe_recurse)
+  if (GDK_IS_PAINTABLE (private->impl))
     {
-      GDK_PAINTABLE_GET_IFACE (private->impl)->invalidate_maybe_recurse (GDK_PAINTABLE (private->impl), region,
-									 child_func, user_data);
+      GdkPaintableIface *iface = GDK_PAINTABLE_GET_IFACE (private->impl);
+
+      if (iface->invalidate_maybe_recurse)
+        iface->invalidate_maybe_recurse ((GdkPaintable*)private->impl, 
+                                         region, child_func, user_data);
       return;
     }
 
