@@ -1,7 +1,7 @@
 /* gdkwindow-quartz.c
  *
  * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
- * Copyright (C) 2005 Imendio AB
+ * Copyright (C) 2005-2007 Imendio AB
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -26,6 +26,13 @@
 
 static gpointer parent_class;
 
+static GSList *update_windows = NULL;
+static guint update_idle = 0;
+
+#define WINDOW_IS_TOPLEVEL(window)		   \
+  (GDK_WINDOW_TYPE (window) != GDK_WINDOW_CHILD && \
+   GDK_WINDOW_TYPE (window) != GDK_WINDOW_FOREIGN)
+
 NSView *
 gdk_quartz_window_get_nsview (GdkWindow *window)
 {
@@ -141,29 +148,97 @@ gdk_window_impl_quartz_begin_paint_region (GdkPaintable *paintable,
 					   GdkRegion    *region)
 {
   GdkWindowImplQuartz *impl = GDK_WINDOW_IMPL_QUARTZ (paintable);
-  CGContextRef context = _gdk_quartz_drawable_get_context (GDK_DRAWABLE (impl), FALSE);
-  int i, n_rects;
+  GdkDrawableImplQuartz *drawable_impl;
+  int n_rects;
   GdkRectangle *rects;
+  GdkPixmap *bg_pixmap;
+  GdkWindow *window;
+
+  drawable_impl = GDK_DRAWABLE_IMPL_QUARTZ (impl);
+  bg_pixmap = GDK_WINDOW_OBJECT (drawable_impl->wrapper)->bg_pixmap;
 
   if (impl->begin_paint_count == 0)
     impl->paint_clip_region = gdk_region_copy (region);
   else
     gdk_region_union (impl->paint_clip_region, region);
 
-  impl->begin_paint_count ++;
+  impl->begin_paint_count++;
+
+  if (bg_pixmap == GDK_NO_BG)
+    return;
 
   gdk_region_get_rectangles (region, &rects, &n_rects);
-  for (i = 0; i < n_rects; i++) 
-    {
-      _gdk_quartz_set_context_fill_color_from_pixel (context, gdk_drawable_get_colormap (GDK_DRAWABLE_IMPL_QUARTZ (impl)->wrapper),
-						     GDK_WINDOW_OBJECT (GDK_DRAWABLE_IMPL_QUARTZ (impl)->wrapper)->bg_color.pixel);
-      
-      CGContextFillRect (context, CGRectMake (rects[i].x, rects[i].y, rects[i].width, rects[i].height));
-      
-    }
-  g_free (rects);
 
-  _gdk_quartz_drawable_release_context (GDK_DRAWABLE (impl), context);
+  if (bg_pixmap == NULL)
+    {
+      CGContextRef cg_context;
+      gfloat r, g, b, a;
+      gint i;
+
+      cg_context = gdk_quartz_drawable_get_context (GDK_DRAWABLE (impl), FALSE);
+      _gdk_quartz_colormap_get_rgba_from_pixel (gdk_drawable_get_colormap (drawable_impl->wrapper),
+				      GDK_WINDOW_OBJECT (drawable_impl->wrapper)->bg_color.pixel,
+				      &r, &g, &b, &a);
+ 
+      for (i = 0; i < n_rects; i++) 
+        {
+	  CGContextSetRGBFillColor (cg_context, r, g, b, a);
+          CGContextFillRect (cg_context,
+			     CGRectMake (rects[i].x, rects[i].y,
+					 rects[i].width, rects[i].height));
+        }
+
+      gdk_quartz_drawable_release_context (GDK_DRAWABLE (impl), cg_context);
+    }
+  else
+    {
+      int x, y;
+      int x_offset, y_offset;
+      int width, height;
+      GdkGC *gc;
+
+      x_offset = y_offset = 0;
+
+      window = GDK_WINDOW (drawable_impl->wrapper);
+      while (window && ((GdkWindowObject *) window)->bg_pixmap == GDK_PARENT_RELATIVE_BG)
+        {
+          /* If this window should have the same background as the parent,
+           * fetch the parent. (And if the same goes for the parent, fetch
+           * the grandparent, etc.)
+           */
+          x_offset += ((GdkWindowObject *) window)->x;
+          y_offset += ((GdkWindowObject *) window)->y;
+          window = GDK_WINDOW (((GdkWindowObject *) window)->parent);
+        }
+
+      /* Note: There should be a CG API to draw tiled images, we might
+       * want to look into that for this. 
+       */
+      gc = gdk_gc_new (GDK_DRAWABLE (impl));
+
+      gdk_drawable_get_size (GDK_DRAWABLE (bg_pixmap), &width, &height);
+
+      x = -x_offset;
+      while (x < (rects[0].x + rects[0].width))
+        {
+          if (x + width >= rects[0].x)
+	    {
+              y = -y_offset;
+              while (y < (rects[0].y + rects[0].height))
+                {
+                  if (y + height >= rects[0].y)
+                    gdk_draw_drawable (GDK_DRAWABLE (impl), gc, bg_pixmap, 0, 0, x, y, width, height);
+		  
+                  y += height;
+                }
+            }
+          x += width;
+        }
+
+      g_object_unref (gc);
+    }
+
+  g_free (rects);
 }
 
 static void
@@ -171,7 +246,7 @@ gdk_window_impl_quartz_end_paint (GdkPaintable *paintable)
 {
   GdkWindowImplQuartz *impl = GDK_WINDOW_IMPL_QUARTZ (paintable);
 
-  impl->begin_paint_count --;
+  impl->begin_paint_count--;
 
   if (impl->begin_paint_count == 0)
     {
@@ -180,27 +255,99 @@ gdk_window_impl_quartz_end_paint (GdkPaintable *paintable)
     }
 }
 
+static void
+gdk_window_quartz_process_updates_internal (GdkWindow *window)
+{
+  GdkWindowObject *private = (GdkWindowObject *) window;
+  GdkWindowImplQuartz *impl = (GdkWindowImplQuartz *) private->impl;
+      
+  if (private->update_area)
+    {
+      int i, n_rects;
+      GdkRectangle *rects;
+
+      gdk_region_get_rectangles (private->update_area, &rects, &n_rects);
+
+      gdk_region_destroy (private->update_area);
+      private->update_area = NULL;
+
+      for (i = 0; i < n_rects; i++) 
+	{
+	  [impl->view setNeedsDisplayInRect:NSMakeRect (rects[i].x, rects[i].y,
+							rects[i].width, rects[i].height)];
+	}
+
+      [impl->view displayIfNeeded];
+
+      g_free (rects);
+    }
+}
+
+static void
+gdk_window_quartz_process_all_updates (void)
+{
+  GSList *old_update_windows = update_windows;
+  GSList *tmp_list = update_windows;
+
+  update_idle = 0;
+  update_windows = NULL;
+
+  g_slist_foreach (old_update_windows, (GFunc) g_object_ref, NULL);
+  
+  GDK_QUARTZ_ALLOC_POOL;
+
+  while (tmp_list)
+    {
+      gdk_window_quartz_process_updates_internal (tmp_list->data);
+
+      g_object_unref (tmp_list->data);
+      tmp_list = tmp_list->next;
+    }
+
+  GDK_QUARTZ_RELEASE_POOL;
+
+  g_slist_free (old_update_windows);
+}
+
+static gboolean
+gdk_window_quartz_update_idle (gpointer data)
+{
+  GDK_THREADS_ENTER ();
+  gdk_window_quartz_process_all_updates ();
+  GDK_THREADS_LEAVE ();
+
+  return FALSE;
+}
+
 static void
 gdk_window_impl_quartz_invalidate_maybe_recurse (GdkPaintable *paintable,
 						 GdkRegion    *region,
 						 gboolean    (*child_func) (GdkWindow *, gpointer),
 						 gpointer      user_data)
 {
-  GdkWindowImplQuartz *impl = GDK_WINDOW_IMPL_QUARTZ (paintable);
-  int i, n_rects;
-  GdkRectangle *rects;
+  GdkWindowImplQuartz *window_impl = GDK_WINDOW_IMPL_QUARTZ (paintable);
+  GdkDrawableImplQuartz *drawable_impl = (GdkDrawableImplQuartz *) window_impl;
+  GdkWindow *window = (GdkWindow *) drawable_impl->wrapper;
+  GdkWindowObject *private = (GdkWindowObject *) window;
+  GdkRegion *visible_region;
 
-  gdk_region_get_rectangles (region, &rects, &n_rects);
+  visible_region = gdk_drawable_get_visible_region (window);
+  gdk_region_intersect (visible_region, region);
 
-  for (i = 0; i < n_rects; i++) 
+  if (private->update_area)
     {
-      [impl->view setNeedsDisplayInRect:NSMakeRect (rects[i].x, rects[i].y,
-						    rects[i].width, rects[i].height)];
+      gdk_region_union (private->update_area, visible_region);
+      gdk_region_destroy (visible_region);
     }
+  else
+    {
+      update_windows = g_slist_prepend (update_windows, window);
+      private->update_area = visible_region;
 
-  g_free (rects);
-
-  /* FIXME: Check if we need to traverse the children */
+      if (update_idle == 0)
+	update_idle = g_idle_add_full (GDK_PRIORITY_REDRAW,
+				       gdk_window_quartz_update_idle, NULL, NULL);
+    }
 }
 
 static void
@@ -208,10 +355,14 @@ gdk_window_impl_quartz_process_updates (GdkPaintable *paintable,
 					gboolean      update_children)
 {
   GdkWindowImplQuartz *impl = GDK_WINDOW_IMPL_QUARTZ (paintable);
+  GdkDrawableImplQuartz *drawable_impl = (GdkDrawableImplQuartz *) impl;
+  GdkWindowObject *private = (GdkWindowObject *) drawable_impl->wrapper;
 
-  [impl->view display];
-
-  /* FIXME: Check if display actually updates the children too */
+  if (private->update_area)
+    {
+      gdk_window_quartz_process_updates_internal ((GdkWindow *) private);
+      update_windows = g_slist_remove (update_windows, private);
+    }
 }
 
 static void
@@ -224,7 +375,6 @@ gdk_window_impl_quartz_paintable_init (GdkPaintableIface *iface)
   iface->process_updates = gdk_window_impl_quartz_process_updates;
 }
 
-
 GType
 _gdk_window_impl_quartz_get_type (void)
 {
@@ -232,25 +382,25 @@ _gdk_window_impl_quartz_get_type (void)
 
   if (!object_type)
     {
-      static const GTypeInfo object_info =
-      {
-        sizeof (GdkWindowImplQuartzClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) gdk_window_impl_quartz_class_init,
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (GdkWindowImplQuartz),
-        0,              /* n_preallocs */
-        (GInstanceInitFunc) gdk_window_impl_quartz_init,
-      };
-      
-      static const GInterfaceInfo paintable_info = 
-      {
-	(GInterfaceInitFunc) gdk_window_impl_quartz_paintable_init,
-	NULL,
-	NULL
-      };
+      const GTypeInfo object_info =
+	{
+	  sizeof (GdkWindowImplQuartzClass),
+	  (GBaseInitFunc) NULL,
+	  (GBaseFinalizeFunc) NULL,
+	  (GClassInitFunc) gdk_window_impl_quartz_class_init,
+	  NULL,           /* class_finalize */
+	  NULL,           /* class_data */
+	  sizeof (GdkWindowImplQuartz),
+	  0,              /* n_preallocs */
+	  (GInstanceInitFunc) gdk_window_impl_quartz_init,
+	};
+
+      const GInterfaceInfo paintable_info = 
+	{
+	  (GInterfaceInitFunc) gdk_window_impl_quartz_paintable_init,
+	  NULL,
+	  NULL
+	};
 
       object_type = g_type_register_static (GDK_TYPE_DRAWABLE_IMPL_QUARTZ,
                                             "GdkWindowImplQuartz",
@@ -259,7 +409,7 @@ _gdk_window_impl_quartz_get_type (void)
 				   GDK_TYPE_PAINTABLE,
 				   &paintable_info);
     }
-  
+
   return object_type;
 }
 
@@ -281,8 +431,9 @@ get_default_title (void)
   return title;
 }
 
+/* FIXME: Should probably be in screen instead. */
 gint 
-_gdk_quartz_get_inverted_screen_y (gint y)
+_gdk_quartz_window_get_inverted_screen_y (gint y)
 {
   NSRect rect = [[NSScreen mainScreen] frame];
 
@@ -291,12 +442,12 @@ _gdk_quartz_get_inverted_screen_y (gint y)
 
 static GdkWindow *
 find_child_window_by_point_helper (GdkWindow *window,
-				   int        x,
-				   int        y,
-				   int        x_offset,
-				   int        y_offset,
-				   int       *x_ret,
-				   int       *y_ret)
+				   gint       x,
+				   gint       y,
+				   gint       x_offset,
+				   gint       y_offset,
+				   gint      *x_ret,
+				   gint      *y_ret)
 {
   GList *l;
 
@@ -336,11 +487,11 @@ find_child_window_by_point_helper (GdkWindow *window,
  * flipped.
  */
 GdkWindow *
-_gdk_quartz_find_child_window_by_point (GdkWindow *toplevel,
-					int        x,
-					int        y,
-					int       *x_ret,
-					int       *y_ret)
+_gdk_quartz_window_find_child_by_point (GdkWindow *toplevel,
+					gint       x,
+					gint       y,
+					gint      *x_ret,
+					gint      *y_ret)
 {
   GdkWindowObject *private = (GdkWindowObject *)toplevel;
   GdkWindowImplQuartz *impl = GDK_WINDOW_IMPL_QUARTZ (private->impl);
@@ -489,9 +640,10 @@ gdk_window_new (GdkWindow     *parent,
     case GDK_WINDOW_DIALOG:
     case GDK_WINDOW_TEMP:
       {
-	NSRect content_rect = NSMakeRect (private->x, 
-					  _gdk_quartz_get_inverted_screen_y (private->y) - impl->height,
-					  impl->width, impl->height);
+	NSRect content_rect = 
+	  NSMakeRect (private->x, 
+		      _gdk_quartz_window_get_inverted_screen_y (private->y) - impl->height,
+		      impl->width, impl->height);
 	const char *title;
 	int style_mask;
 
@@ -508,15 +660,14 @@ gdk_window_new (GdkWindow     *parent,
 
 	impl->toplevel = [[GdkQuartzWindow alloc] initWithContentRect:content_rect 
 			                                    styleMask:style_mask
-			                                      backing:NSBackingStoreBuffered defer:NO];
+			                                      backing:NSBackingStoreBuffered
+			                                        defer:NO];
+
 	if (attributes_mask & GDK_WA_TITLE)
 	  title = attributes->title;
 	else
 	  title = get_default_title ();
 
-	if (attributes->window_type == GDK_WINDOW_TEMP)
-	  [impl->toplevel setLevel:NSPopUpMenuWindowLevel];
-
 	gdk_window_set_title (window, title);
 	  
 	if (draw_impl->colormap == gdk_screen_get_rgba_colormap (_gdk_screen))
@@ -560,6 +711,9 @@ gdk_window_new (GdkWindow     *parent,
 
   GDK_QUARTZ_RELEASE_POOL;
 
+  if (attributes_mask & GDK_WA_TYPE_HINT)
+    gdk_window_set_type_hint (window, attributes->type_hint);
+
   return window;
 }
 
@@ -584,20 +738,25 @@ _gdk_windowing_window_destroy (GdkWindow *window,
 			       gboolean   recursing,
 			       gboolean   foreign_destroy)
 {
-  if (!recursing && ! foreign_destroy)
+  update_windows = g_slist_remove (update_windows, window);
+
+  if (!recursing && !foreign_destroy)
     {
       GdkWindowImplQuartz *impl = GDK_WINDOW_IMPL_QUARTZ (GDK_WINDOW_OBJECT (window)->impl);
 
-      if (window == _gdk_quartz_get_mouse_window ()) 
-	{
-	  _gdk_quartz_update_mouse_window (_gdk_root);
-	}
+      if (window == _gdk_quartz_events_get_mouse_window ()) 
+	_gdk_quartz_events_update_mouse_window (_gdk_root);
+
+      GDK_QUARTZ_ALLOC_POOL;
+
+      _gdk_quartz_drawable_finish (GDK_DRAWABLE (impl));
 
       if (impl->toplevel)
 	[impl->toplevel close];
       else if (impl->view)
 	[impl->view release];
-      
+
+      GDK_QUARTZ_RELEASE_POOL;
     }
 }
 
@@ -622,7 +781,8 @@ all_parents_shown (GdkWindowObject *private)
 }
 
 static void
-show_window_internal (GdkWindow *window, gboolean raise)
+show_window_internal (GdkWindow *window,
+		      gboolean   raise)
 {
   GdkWindowObject *private;
   GdkWindowImplQuartz *impl;
@@ -635,8 +795,6 @@ show_window_internal (GdkWindow *window, gboolean raise)
   private = (GdkWindowObject *)window;
   impl = GDK_WINDOW_IMPL_QUARTZ (private->impl);
 
-  /* FIXME: We need to raise the window (move it to the top in the list) */
-
   if (impl->toplevel)
     {
       [impl->toplevel orderFront:nil];
@@ -649,10 +807,16 @@ show_window_internal (GdkWindow *window, gboolean raise)
     }
 
   if (all_parents_shown (private->parent))
-    _gdk_quartz_send_map_events (window);
+    _gdk_quartz_events_send_map_events (window);
 
   gdk_synthesize_window_state (window, GDK_WINDOW_STATE_WITHDRAWN, 0);
 
+  if (private->state & GDK_WINDOW_STATE_MAXIMIZED)
+    gdk_window_maximize (window);
+
+  if (private->state & GDK_WINDOW_STATE_ICONIFIED)
+    gdk_window_iconify (window);
+
   GDK_QUARTZ_RELEASE_POOL;
 }
 
@@ -745,9 +909,10 @@ move_resize_window_internal (GdkWindow *window,
 
   if (impl->toplevel)
     {
-      NSRect content_rect = NSMakeRect (private->x, 
-					_gdk_quartz_get_inverted_screen_y (private->y) ,
-					impl->width, impl->height);
+      NSRect content_rect = 
+	NSMakeRect (private->x, 
+		    _gdk_quartz_window_get_inverted_screen_y (private->y) ,
+		    impl->width, impl->height);
       NSRect frame_rect = [impl->toplevel frameRectForContentRect:content_rect];
       
       frame_rect.origin.y -= frame_rect.size.height;
@@ -758,10 +923,10 @@ move_resize_window_internal (GdkWindow *window,
     {
       if (!private->input_only)
 	{
-	  [impl->view setFrame:NSMakeRect (private->x, private->y, 
-					   impl->width, impl->height)];
+	  NSRect nsrect;
 
-	  /* FIXME: Maybe we should use setNeedsDisplayInRect instead */
+	  nsrect = NSMakeRect (private->x, private->y, impl->width, impl->height);
+	  [impl->view setFrame: nsrect];
 	  [impl->view setNeedsDisplay:YES];
 	}
     }
@@ -945,7 +1110,7 @@ gdk_window_set_cursor (GdkWindow *window,
 
   impl->nscursor = nscursor;
 
-  _gdk_quartz_update_cursor (_gdk_quartz_get_mouse_window ());
+  _gdk_quartz_events_update_cursor (_gdk_quartz_events_get_mouse_window ());
 }
 
 void
@@ -992,7 +1157,7 @@ gdk_window_get_origin (GdkWindow *window,
   content_rect = [impl->toplevel contentRectForFrameRect:[impl->toplevel frame]];
 
   tmp_x = content_rect.origin.x;
-  tmp_y = _gdk_quartz_get_inverted_screen_y (content_rect.origin.y + content_rect.size.height);
+  tmp_y = _gdk_quartz_window_get_inverted_screen_y (content_rect.origin.y + content_rect.size.height);
 
   while (private != GDK_WINDOW_OBJECT (toplevel))
     {
@@ -1084,13 +1249,15 @@ _gdk_windowing_window_get_pointer (GdkDisplay      *display,
   if (window == _gdk_root)
     {
       point = [NSEvent mouseLocation];
-      y_tmp = _gdk_quartz_get_inverted_screen_y (point.y);
+      y_tmp = _gdk_quartz_window_get_inverted_screen_y (point.y);
+      *mask = _gdk_quartz_events_get_current_event_mask ();
     }
   else
     {
       NSWindow *nswindow = impl->toplevel;
       point = [nswindow mouseLocationOutsideOfEventStream];
       y_tmp = impl->height - point.y;
+      *mask = _gdk_quartz_events_get_current_event_mask ();
     }
   x_tmp = point.x;
 
@@ -1102,12 +1269,10 @@ _gdk_windowing_window_get_pointer (GdkDisplay      *display,
       private = private->parent;
     }
 
-  /* FIXME: Implement mask. */
   *x = x_tmp;
   *y = y_tmp;
-  *mask = 0;
 
-  return _gdk_quartz_find_child_window_by_point (window,
+  return _gdk_quartz_window_find_child_by_point (window,
 						 point.x, point.y,
 						 &x_tmp, &y_tmp);
 }
@@ -1169,7 +1334,77 @@ gdk_window_set_geometry_hints (GdkWindow      *window,
 			       GdkGeometry    *geometry,
 			       GdkWindowHints  geom_mask)
 {
-  /* FIXME: Implement */
+  GdkWindowImplQuartz *impl;
+
+  g_return_if_fail (GDK_IS_WINDOW (window));
+  g_return_if_fail (geometry != NULL);
+
+  if (GDK_WINDOW_DESTROYED (window))
+    return;
+  
+  impl = GDK_WINDOW_IMPL_QUARTZ (((GdkWindowObject *) window)->impl);
+  if (!impl->toplevel)
+    return;
+
+  if (geom_mask & GDK_HINT_POS)
+    {
+      /* FIXME: Implement */
+    }
+
+  if (geom_mask & GDK_HINT_USER_POS)
+    {
+      /* FIXME: Implement */
+    }
+
+  if (geom_mask & GDK_HINT_USER_SIZE)
+    {
+      /* FIXME: Implement */
+    }
+  
+  if (geom_mask & GDK_HINT_MIN_SIZE)
+    {
+      NSSize size;
+
+      size.width = geometry->min_width;
+      size.height = geometry->min_height;
+
+      [impl->toplevel setContentMinSize:size];
+    }
+  
+  if (geom_mask & GDK_HINT_MAX_SIZE)
+    {
+      NSSize size;
+
+      size.width = geometry->max_width;
+      size.height = geometry->max_height;
+
+      [impl->toplevel setContentMaxSize:size];
+    }
+  
+  if (geom_mask & GDK_HINT_BASE_SIZE)
+    {
+      /* FIXME: Implement */
+    }
+  
+  if (geom_mask & GDK_HINT_RESIZE_INC)
+    {
+      NSSize size;
+
+      size.width = geometry->width_inc;
+      size.height = geometry->height_inc;
+
+      [impl->toplevel setContentResizeIncrements:size];
+    }
+  
+  if (geom_mask & GDK_HINT_ASPECT)
+    {
+      /* FIXME: Implement */
+    }
+
+  if (geom_mask & GDK_HINT_WIN_GRAVITY)
+    {
+      /* FIXME: Implement */
+    }
 }
 
 void
@@ -1201,7 +1436,7 @@ gdk_window_set_role (GdkWindow   *window,
   /* FIXME: Implement */
 }
 
-void          
+void
 gdk_window_set_transient_for (GdkWindow *window, 
 			      GdkWindow *parent)
 {
@@ -1308,7 +1543,6 @@ gdk_window_set_icon (GdkWindow *window,
 		     GdkPixmap *pixmap,
 		     GdkBitmap *mask)
 {
-  g_return_if_fail (window != NULL);
   g_return_if_fail (GDK_IS_WINDOW (window));
 
   /* FIXME: Implement */
@@ -1318,7 +1552,6 @@ void
 gdk_window_set_icon_name (GdkWindow   *window, 
 			  const gchar *name)
 {
-  g_return_if_fail (window != NULL);
   g_return_if_fail (GDK_IS_WINDOW (window));
 
   /* FIXME: Implement */
@@ -1350,12 +1583,83 @@ void
 gdk_window_set_type_hint (GdkWindow        *window,
 			  GdkWindowTypeHint hint)
 {
+  GdkWindowImplQuartz *impl;
+  gint                 level;
+  gboolean             shadow;
+  
   g_return_if_fail (GDK_IS_WINDOW (window));
 
   if (GDK_WINDOW_DESTROYED (window))
     return;
 
-  GDK_WINDOW_IMPL_QUARTZ (((GdkWindowObject *) window)->impl)->type_hint = hint;
+  impl = GDK_WINDOW_IMPL_QUARTZ (((GdkWindowObject *) window)->impl);
+
+  impl->type_hint = hint;
+
+  /* Match the documentation, only do something if we're not mapped yet. */
+  if (GDK_WINDOW_IS_MAPPED (window))
+    return;
+
+  switch (hint)
+    {
+    case GDK_WINDOW_TYPE_HINT_NORMAL:  /* Normal toplevel window */
+    case GDK_WINDOW_TYPE_HINT_DIALOG:  /* Dialog window */
+      level = NSNormalWindowLevel;
+      shadow = TRUE;
+      break;
+
+    case GDK_WINDOW_TYPE_HINT_TOOLBAR: /* Window used to implement toolbars */
+    case GDK_WINDOW_TYPE_HINT_DESKTOP: /* N/A */
+      level = NSNormalWindowLevel;
+      shadow = FALSE;
+      break;
+
+    case GDK_WINDOW_TYPE_HINT_DOCK:
+    case GDK_WINDOW_TYPE_HINT_UTILITY:
+      level = NSFloatingWindowLevel;
+      shadow = TRUE;
+      break;
+
+    case GDK_WINDOW_TYPE_HINT_MENU: /* Torn-off menu */
+      level = NSTornOffMenuWindowLevel;
+      shadow = TRUE;
+      break;
+      
+    case GDK_WINDOW_TYPE_HINT_DROPDOWN_MENU: /* Menu from menubar */
+      level = NSTornOffMenuWindowLevel;
+      shadow = TRUE;
+      break;
+
+    case GDK_WINDOW_TYPE_HINT_SPLASHSCREEN:
+      level = NSPopUpMenuWindowLevel;
+      shadow = TRUE;
+      break;
+
+    case GDK_WINDOW_TYPE_HINT_POPUP_MENU:
+    case GDK_WINDOW_TYPE_HINT_COMBO:
+      level = NSPopUpMenuWindowLevel;
+      shadow = TRUE;
+      break;
+      
+    case GDK_WINDOW_TYPE_HINT_NOTIFICATION:
+    case GDK_WINDOW_TYPE_HINT_TOOLTIP:
+      level = NSStatusWindowLevel;
+      shadow = TRUE;
+      break;
+
+    case GDK_WINDOW_TYPE_HINT_DND:
+      level = NSPopUpMenuWindowLevel;
+      shadow = FALSE;
+      break;
+
+    default:
+      level = NSNormalWindowLevel;
+      shadow = FALSE;
+      break;
+    }
+
+  [impl->toplevel setHasShadow:shadow];
+  [impl->toplevel setLevel:level];
 }
 
 GdkWindowTypeHint
@@ -1454,7 +1758,7 @@ gdk_window_get_frame_extents (GdkWindow    *window,
   ns_rect = [impl->toplevel frame];
 
   rect->x = ns_rect.origin.x;
-  rect->y = _gdk_quartz_get_inverted_screen_y (ns_rect.origin.y + ns_rect.size.height);
+  rect->y = _gdk_quartz_window_get_inverted_screen_y (ns_rect.origin.y + ns_rect.size.height);
   rect->width = ns_rect.size.width;
   rect->height = ns_rect.size.height;
 }
@@ -1515,17 +1819,59 @@ gdk_window_unstick (GdkWindow *window)
 void
 gdk_window_maximize (GdkWindow *window)
 {
+  GdkWindowImplQuartz *impl;
+
   g_return_if_fail (GDK_IS_WINDOW (window));
 
-  /* FIXME: Implement */
+  if (GDK_WINDOW_DESTROYED (window))
+    return;
+
+  impl = GDK_WINDOW_IMPL_QUARTZ (GDK_WINDOW_OBJECT (window)->impl);
+
+  if (GDK_WINDOW_IS_MAPPED (window))
+    {
+      GDK_QUARTZ_ALLOC_POOL;
+
+      if (impl->toplevel && ![impl->toplevel isZoomed])
+	[impl->toplevel zoom:nil];
+
+      GDK_QUARTZ_RELEASE_POOL;
+    }
+  else
+    {
+      gdk_synthesize_window_state (window,
+				   0,
+				   GDK_WINDOW_STATE_MAXIMIZED);
+    }
 }
 
 void
 gdk_window_unmaximize (GdkWindow *window)
 {
+  GdkWindowImplQuartz *impl;
+
   g_return_if_fail (GDK_IS_WINDOW (window));
 
-  /* FIXME: Implement */
+  if (GDK_WINDOW_DESTROYED (window))
+    return;
+
+  impl = GDK_WINDOW_IMPL_QUARTZ (GDK_WINDOW_OBJECT (window)->impl);
+
+  if (GDK_WINDOW_IS_MAPPED (window))
+    {
+      GDK_QUARTZ_ALLOC_POOL;
+
+      if (impl->toplevel && [impl->toplevel isZoomed])
+	[impl->toplevel zoom:nil];
+
+      GDK_QUARTZ_RELEASE_POOL;
+    }
+  else
+    {
+      gdk_synthesize_window_state (window,
+				   GDK_WINDOW_STATE_MAXIMIZED,
+				   0);
+    }
 }
 
 void
@@ -1539,12 +1885,22 @@ gdk_window_iconify (GdkWindow *window)
     return;
   
   impl = GDK_WINDOW_IMPL_QUARTZ (GDK_WINDOW_OBJECT (window)->impl);
-  if (!impl->toplevel)
-    return;
-  
-  GDK_QUARTZ_ALLOC_POOL;
-  [impl->toplevel miniaturize:nil];
-  GDK_QUARTZ_RELEASE_POOL;
+
+  if (GDK_WINDOW_IS_MAPPED (window))
+    {
+      GDK_QUARTZ_ALLOC_POOL;
+
+      if (impl->toplevel)
+	[impl->toplevel miniaturize:nil];
+
+      GDK_QUARTZ_RELEASE_POOL;
+    }
+  else
+    {
+      gdk_synthesize_window_state (window,
+				   0,
+				   GDK_WINDOW_STATE_ICONIFIED);
+    }
 }
 
 void
@@ -1558,12 +1914,22 @@ gdk_window_deiconify (GdkWindow *window)
     return;
 
   impl = GDK_WINDOW_IMPL_QUARTZ (GDK_WINDOW_OBJECT (window)->impl);
-  if (!impl->toplevel)
-    return;
 
-  GDK_QUARTZ_ALLOC_POOL;
-  [impl->toplevel deminiaturize:nil];
-  GDK_QUARTZ_RELEASE_POOL;
+  if (GDK_WINDOW_IS_MAPPED (window))
+    {
+      GDK_QUARTZ_ALLOC_POOL;
+
+      if (impl->toplevel)
+	[impl->toplevel deminiaturize:nil];
+
+      GDK_QUARTZ_RELEASE_POOL;
+    }
+  else
+    {
+      gdk_synthesize_window_state (window,
+				   GDK_WINDOW_STATE_ICONIFIED,
+				   0);
+    }
 }
 
 void
@@ -1655,3 +2021,30 @@ gdk_window_destroy_notify (GdkWindow *window)
    * -[GdkQuartzView dealloc], although I suspect that currently they leak
    * anyway. */
 }
+
+void 
+gdk_window_beep (GdkWindow *window)
+{
+  gdk_display_beep (_gdk_display);
+}
+
+void
+gdk_window_set_opacity (GdkWindow *window,
+			gdouble    opacity)
+{
+  GdkWindowObject *private = (GdkWindowObject *) window;
+  GdkWindowImplQuartz *impl = GDK_WINDOW_IMPL_QUARTZ (private->impl);
+
+  g_return_if_fail (GDK_IS_WINDOW (window));
+  g_return_if_fail (WINDOW_IS_TOPLEVEL (window));
+
+  if (GDK_WINDOW_DESTROYED (window))
+    return;
+
+  if (opacity < 0)
+    opacity = 0;
+  else if (opacity > 1)
+    opacity = 1;
+
+  [impl->toplevel setAlphaValue: opacity];
+}
