@@ -43,12 +43,6 @@
 
 #include <glib/gprintf.h>
 
-#if defined (__GNUC__) && defined (HAVE_DIMM_H)
-/* The w32api imm.h clashes a bit with the IE5.5 dimm.h */
-# define IMEMENUITEMINFOA hidden_IMEMENUITEMINFOA
-# define IMEMENUITEMINFOW hidden_IMEMENUITEMINFOW
-#endif
-
 #include "gdk.h"
 #include "gdkprivate-win32.h"
 #include "gdkinput-win32.h"
@@ -65,15 +59,6 @@
 
 #include <imm.h>
 
-#if defined (__GNUC__) && defined (HAVE_DIMM_H)
-# undef IMEMENUITEMINFOA
-# undef IMEMENUITEMINFOW
-#endif
-
-#ifdef HAVE_DIMM_H
-#include <dimm.h>
-#endif
-
 #ifndef XBUTTON1
 #define XBUTTON1 1
 #define XBUTTON2 2
@@ -144,10 +129,8 @@ static gint current_root_x, current_root_y;
 static UINT msh_mousewheel;
 static UINT client_message;
 
-#ifdef HAVE_DIMM_H
-static IActiveIMMApp *active_imm_app = NULL;
-static IActiveIMMMessagePumpOwner *active_imm_msgpump_owner = NULL;
-#endif
+static UINT got_gdk_events_message;
+static HWND modal_win32_dialog = NULL;
 
 #if 0
 static HKL latin_locale = NULL;
@@ -262,9 +245,6 @@ inner_window_procedure (HWND   hwnd,
 {
   MSG msg;
   DWORD pos;
-#ifdef HAVE_DIMM_H
-  LRESULT lres;
-#endif
   gint ret_val = 0;
 
   msg.hwnd = hwnd;
@@ -281,21 +261,16 @@ inner_window_procedure (HWND   hwnd,
       /* If gdk_event_translate() returns TRUE, we return ret_val from
        * the window procedure.
        */
+      if (modal_win32_dialog)
+	PostMessageW (modal_win32_dialog, got_gdk_events_message,
+		      (WPARAM) 1, 0);
       return ret_val;
     }
   else
     {
-      /* Otherwise call DefWindowProc(). */
-      GDK_NOTE (EVENTS, g_print (" DefWindowProc"));
-#ifndef HAVE_DIMM_H
-      return DefWindowProc (hwnd, message, wparam, lparam);
-#else
-      if (active_imm_app == NULL ||
-	  (*active_imm_app->lpVtbl->OnDefWindowProc) (active_imm_app, hwnd, message, wparam, lparam, &lres) == S_FALSE)
-	return DefWindowProc (hwnd, message, wparam, lparam);
-      else
-	return lres;
-#endif
+      /* Otherwise call DefWindowProcW(). */
+      GDK_NOTE (EVENTS, g_print (" DefWindowProcW"));
+      return DefWindowProcW (hwnd, message, wparam, lparam);
     }
 }
 
@@ -325,9 +300,6 @@ void
 _gdk_events_init (void)
 {
   GSource *source;
-#ifdef HAVE_DIMM_H
-  HRESULT hres;
-#endif
 
 #if 0
   int i, j, n;
@@ -379,6 +351,7 @@ _gdk_events_init (void)
   msh_mousewheel = RegisterWindowMessage ("MSWHEEL_ROLLMSG");
 
   client_message = RegisterWindowMessage ("GDK_WIN32_CLIENT_MESSAGE");
+  got_gdk_events_message = RegisterWindowMessage ("GDK_WIN32_GOT_EVENTS");
 
 #if 0
   /* Check if we have some input locale identifier loaded that uses a
@@ -436,26 +409,6 @@ _gdk_events_init (void)
   g_source_add_poll (source, &event_poll_fd);
   g_source_set_can_recurse (source, TRUE);
   g_source_attach (source, NULL);
-
-#ifdef HAVE_DIMM_H
-  hres = CoCreateInstance (&CLSID_CActiveIMM,
-			   NULL,
-			   CLSCTX_ALL,
-			   &IID_IActiveIMMApp,
-			   (LPVOID *) &active_imm_app);
-  
-  if (hres == S_OK)
-    {
-      GDK_NOTE (EVENTS, g_print ("IActiveIMMApp created %p\n",
-				 active_imm_app));
-      (*active_imm_app->lpVtbl->Activate) (active_imm_app, TRUE);
-      
-      hres = (*active_imm_app->lpVtbl->QueryInterface) (active_imm_app, &IID_IActiveIMMMessagePumpOwner, (void **) &active_imm_msgpump_owner);
-      GDK_NOTE (EVENTS, g_print ("IActiveIMMMessagePumpOwner created %p\n",
-				 active_imm_msgpump_owner));
-      (active_imm_msgpump_owner->lpVtbl->Start) (active_imm_msgpump_owner);
-    }
-#endif
 }
 
 gboolean
@@ -463,7 +416,8 @@ gdk_events_pending (void)
 {
   MSG msg;
   return (_gdk_event_queue_find_first (_gdk_display) ||
-	  PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE));
+	  (modal_win32_dialog == NULL &&
+	   PeekMessageW (&msg, NULL, 0, 0, PM_NOREMOVE)));
 }
 
 GdkEvent*
@@ -476,7 +430,7 @@ gdk_event_get_graphics_expose (GdkWindow *window)
   
   GDK_NOTE (EVENTS, g_print ("gdk_event_get_graphics_expose\n"));
 
-  if (PeekMessage (&msg, GDK_WINDOW_HWND (window), WM_PAINT, WM_PAINT, PM_REMOVE))
+  if (PeekMessageW (&msg, GDK_WINDOW_HWND (window), WM_PAINT, WM_PAINT, PM_REMOVE))
     {
       handle_wm_paint (&msg, window, TRUE, &event);
       if (event != NULL)
@@ -585,7 +539,7 @@ gdk_pointer_grab (GdkWindow    *window,
       if (p_grab_window != NULL && p_grab_window != window)
 	generate_grab_broken_event (p_grab_window, FALSE, window);
       
-      p_grab_window = window;
+      assign_object (&p_grab_window, window);
 
       if (p_grab_cursor != NULL)
 	{
@@ -642,7 +596,7 @@ gdk_display_pointer_ungrab (GdkDisplay *display,
 
   /* FIXME: Generate GDK_CROSSING_UNGRAB events */
 
-  p_grab_window = NULL;
+  assign_object (&p_grab_window, NULL);
   if (p_grab_cursor != NULL)
     {
       if (GetCursor () == p_grab_cursor)
@@ -759,7 +713,7 @@ gdk_keyboard_grab (GdkWindow *window,
       if (k_grab_window != NULL && k_grab_window != window)
 	generate_grab_broken_event (k_grab_window, TRUE, window);
 
-      k_grab_window = window;
+      assign_object (&k_grab_window, window);
 
       if (!k_grab_owner_events)
 	{
@@ -824,7 +778,7 @@ gdk_display_keyboard_ungrab (GdkDisplay *display,
 	}
     }
 
-  k_grab_window = NULL;
+  assign_object (&k_grab_window, NULL);
 }
 
 gboolean
@@ -894,21 +848,6 @@ build_key_event_state (GdkEvent *event,
   if (key_state[VK_XBUTTON2] & 0x80)
     event->key.state |= GDK_BUTTON5_MASK;
 
-  /* Win9x doesn't distinguish between left and right Control and Alt
-   * in the keyboard state as returned by GetKeyboardState(), so we
-   * have to punt, and accept either Control + either Alt to be AltGr.
-   *
-   * Alternatively, we could have some state saved when the Control
-   * and Alt keys messages come in, as the KF_EXTENDED bit in lParam
-   * does indicate correctly whether it is the right Control or Alt
-   * key. But that would be a bit messy.
-   */
-  if (!G_WIN32_IS_NT_BASED () &&
-      _gdk_keyboard_has_altgr &&
-      key_state[VK_CONTROL] & 0x80 &&
-      key_state[VK_MENU] & 0x80)
-    key_state[VK_LCONTROL] = key_state[VK_RMENU] = 0x80;
-
   if (_gdk_keyboard_has_altgr &&
       (key_state[VK_LCONTROL] & 0x80) &&
       (key_state[VK_RMENU] & 0x80))
@@ -1793,11 +1732,12 @@ static gboolean
 doesnt_want_scroll (gint mask,
 		    MSG *msg)
 {
-#if 0
-  return !(mask & GDK_SCROLL_MASK);
-#else
-  return !(mask & GDK_BUTTON_PRESS_MASK);
-#endif
+  /* As there are no separate scroll events in X11, button press
+   * events are used, so higher level code might be selecting for
+   * either GDK_BUTTON_PRESS_MASK or GDK_SCROLL_MASK when it wants GDK
+   * scroll events. Make sure this works in the Win32 backend, too.
+   */
+  return !(mask & (GDK_SCROLL_MASK|GDK_BUTTON_PRESS_MASK));
 }
 
 static void
@@ -1840,139 +1780,6 @@ handle_configure_event (MSG       *msg,
     }
 }
 
-static void
-erase_background (GdkWindow *window,
-		  HDC        hdc)
-{
-  HDC bgdc = NULL;
-  HBRUSH hbr = NULL;
-  HPALETTE holdpal = NULL;
-  RECT rect;
-  COLORREF bg;
-  GdkColormap *colormap;
-  GdkColormapPrivateWin32 *colormap_private;
-  int x, y;
-  int x_offset, y_offset;
-  
-  if (((GdkWindowObject *) window)->input_only ||
-      ((GdkWindowObject *) window)->bg_pixmap == GDK_NO_BG ||
-      GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->position_info.no_bg)
-    {
-      return;
-    }
-
-  colormap = gdk_drawable_get_colormap (window);
-
-  if (colormap &&
-      (colormap->visual->type == GDK_VISUAL_PSEUDO_COLOR ||
-       colormap->visual->type == GDK_VISUAL_STATIC_COLOR))
-    {
-      int k;
-	  
-      colormap_private = GDK_WIN32_COLORMAP_DATA (colormap);
-
-      if (!(holdpal = SelectPalette (hdc,  colormap_private->hpal, FALSE)))
-        WIN32_GDI_FAILED ("SelectPalette");
-      else if ((k = RealizePalette (hdc)) == GDI_ERROR)
-	WIN32_GDI_FAILED ("RealizePalette");
-      else if (k > 0)
-	GDK_NOTE (COLORMAP, g_print ("erase_background: realized %p: %d colors\n",
-				     colormap_private->hpal, k));
-    }
-  
-  x_offset = y_offset = 0;
-  while (window && ((GdkWindowObject *) window)->bg_pixmap == GDK_PARENT_RELATIVE_BG)
-    {
-      /* If this window should have the same background as the parent,
-       * fetch the parent. (And if the same goes for the parent, fetch
-       * the grandparent, etc.)
-       */
-      x_offset += ((GdkWindowObject *) window)->x;
-      y_offset += ((GdkWindowObject *) window)->y;
-      window = GDK_WINDOW (((GdkWindowObject *) window)->parent);
-    }
-  
-  if (GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->position_info.no_bg)
-    {
-      /* Improves scolling effect, e.g. main buttons of testgtk */
-      return;
-    }
-
-  GetClipBox (hdc, &rect);
-
-  if (((GdkWindowObject *) window)->bg_pixmap == NULL)
-    {
-      bg = _gdk_win32_colormap_color (GDK_DRAWABLE_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->colormap,
-				      ((GdkWindowObject *) window)->bg_color.pixel);
-      
-      if (!(hbr = CreateSolidBrush (bg)))
-	WIN32_GDI_FAILED ("CreateSolidBrush");
-      else if (!FillRect (hdc, &rect, hbr))
-	WIN32_GDI_FAILED ("FillRect");
-      if (hbr != NULL)
-	DeleteObject (hbr);
-    }
-  else if (((GdkWindowObject *) window)->bg_pixmap != GDK_NO_BG)
-    {
-      GdkPixmap *pixmap = ((GdkWindowObject *) window)->bg_pixmap;
-      GdkPixmapImplWin32 *pixmap_impl = GDK_PIXMAP_IMPL_WIN32 (GDK_PIXMAP_OBJECT (pixmap)->impl);
-      
-      if (x_offset == 0 && y_offset == 0 &&
-	  pixmap_impl->width <= 8 && pixmap_impl->height <= 8)
-	{
-	  if (!(hbr = CreatePatternBrush (GDK_PIXMAP_HBITMAP (pixmap))))
-	    WIN32_GDI_FAILED ("CreatePatternBrush");
-	  else if (!FillRect (hdc, &rect, hbr))
-	    WIN32_GDI_FAILED ("FillRect");
-	  if (hbr != NULL)
-	    DeleteObject (hbr);
-	}
-      else
-	{
-	  HGDIOBJ oldbitmap;
-
-	  if (!(bgdc = CreateCompatibleDC (hdc)))
-	    {
-	      WIN32_GDI_FAILED ("CreateCompatibleDC");
-	      return;
-	    }
-	  if (!(oldbitmap = SelectObject (bgdc, GDK_PIXMAP_HBITMAP (pixmap))))
-	    {
-	      WIN32_GDI_FAILED ("SelectObject");
-	      DeleteDC (bgdc);
-	      return;
-	    }
-	  x = -x_offset;
-	  while (x < rect.right)
-	    {
-	      if (x + pixmap_impl->width >= rect.left)
-		{
-		  y = -y_offset;
-		  while (y < rect.bottom)
-		    {
-		      if (y + pixmap_impl->height >= rect.top)
-			{
-			  if (!BitBlt (hdc, x, y,
-				       pixmap_impl->width, pixmap_impl->height,
-				       bgdc, 0, 0, SRCCOPY))
-			    {
-			      WIN32_GDI_FAILED ("BitBlt");
-			      SelectObject (bgdc, oldbitmap);
-			      DeleteDC (bgdc);
-			      return;
-			    }
-			}
-		      y += pixmap_impl->height;
-		    }
-		}
-	      x += pixmap_impl->width;
-	    }
-	  SelectObject (bgdc, oldbitmap);
-	  DeleteDC (bgdc);
-	}
-    }
-}
-
 GdkRegion *
 _gdk_win32_hrgn_to_region (HRGN hrgn)
 {
@@ -2132,8 +1939,8 @@ static void
 handle_stuff_while_moving_or_resizing (void)
 {
   int arbitrary_limit = 1;
-  while (g_main_pending () && arbitrary_limit--)
-    g_main_iteration (FALSE);
+  while (g_main_context_pending (NULL) && arbitrary_limit--)
+    g_main_context_iteration (NULL, FALSE);
 }
 
 static VOID CALLBACK
@@ -2226,7 +2033,7 @@ gdk_event_translate (MSG  *msg,
       
       /* If result is GDK_FILTER_CONTINUE, we continue as if nothing
        * happened. If it is GDK_FILTER_REMOVE or GDK_FILTER_TRANSLATE,
-       * we return TRUE, and DefWindowProc() will not be called.
+       * we return TRUE, and DefWindowProcW() will not be called.
        */
       if (result == GDK_FILTER_REMOVE || result == GDK_FILTER_TRANSLATE)
 	return TRUE;
@@ -2254,12 +2061,11 @@ gdk_event_translate (MSG  *msg,
 	   */
 	  GDK_NOTE (EVENTS, g_print (" (posted)"));
 	
-	  PostMessage (msg->hwnd, msg->message,
-		       msg->wParam, msg->lParam);
+	  PostMessageW (msg->hwnd, msg->message, msg->wParam, msg->lParam);
 	}
       else if (msg->message == WM_CREATE)
 	{
-	  window = (UNALIGNED GdkWindow*) (((LPCREATESTRUCT) msg->lParam)->lpCreateParams);
+	  window = (UNALIGNED GdkWindow*) (((LPCREATESTRUCTW) msg->lParam)->lpCreateParams);
 	  GDK_WINDOW_HWND (window) = msg->hwnd;
 	}
       else
@@ -2513,7 +2319,7 @@ gdk_event_translate (MSG  *msg,
     case WM_SYSCHAR:
       if (msg->wParam != VK_SPACE)
 	{
-	  /* To prevent beeps, don't let DefWindowProc() be called */
+	  /* To prevent beeps, don't let DefWindowProcW() be called */
 	  return_val = TRUE;
 	  goto done;
 	}
@@ -2749,7 +2555,22 @@ gdk_event_translate (MSG  *msg,
 	  GdkWindow *real_window = find_real_window_for_grabbed_mouse_event (window, msg);
 
 	  if (real_window != current_window)
-	    synthesize_crossing_events (real_window, GDK_CROSSING_NORMAL, msg);
+	    {
+	      if (p_grab_owner_events)
+		{
+		  synthesize_crossing_events (real_window, GDK_CROSSING_NORMAL, msg);
+		}
+	      else if (current_window == p_grab_window)
+		{
+		  synthesize_leave_event (p_grab_window, msg, GDK_CROSSING_NORMAL, GDK_NOTIFY_ANCESTOR);
+		  assign_object (&current_window, _gdk_root);
+		}
+	      else if (real_window == p_grab_window)
+		{
+		  synthesize_enter_event (p_grab_window, msg, GDK_CROSSING_NORMAL, GDK_NOTIFY_ANCESTOR);
+		  assign_object (&current_window, p_grab_window);
+		}
+	    }
 	}
       else
 	{
@@ -2817,6 +2638,7 @@ gdk_event_translate (MSG  *msg,
 	  synthesize_enter_or_leave_event (current_window, msg, 
 					   GDK_LEAVE_NOTIFY, GDK_CROSSING_NORMAL, detail,
 					   current_x, current_y);
+	  assign_object (&current_window, _gdk_root);
 	}
       else
 	{
@@ -2906,7 +2728,8 @@ gdk_event_translate (MSG  *msg,
       break;
 
      case WM_MOUSEACTIVATE:
-       if (gdk_window_get_window_type (window) == GDK_WINDOW_TEMP) 
+       if (gdk_window_get_window_type (window) == GDK_WINDOW_TEMP 
+	   || !((GdkWindowObject *)window)->accept_focus)
 	 {
 	   *ret_valp = MA_NOACTIVATE;
 	   return_val = TRUE;
@@ -2942,7 +2765,6 @@ gdk_event_translate (MSG  *msg,
       if (GDK_WINDOW_DESTROYED (window))
 	break;
 
-      erase_background (window, (HDC) msg->wParam);
       return_val = TRUE;
       *ret_valp = 1;
       break;
@@ -3077,55 +2899,14 @@ gdk_event_translate (MSG  *msg,
 	 GDK_WINDOW_TYPE (window) != GDK_WINDOW_CHILD &&
 	 !GDK_WINDOW_DESTROYED (window))
 	{
-	  RECT client_rect;
-	  POINT point;
-
-	  GetClientRect (msg->hwnd, &client_rect);
-	  point.x = client_rect.left; /* always 0 */
-	  point.y = client_rect.top;
-	  /* top level windows need screen coords */
-	  if (gdk_window_get_parent (window) == _gdk_root)
-	    {
-	      ClientToScreen (msg->hwnd, &point);
-	      point.x += _gdk_offset_x;
-	      point.y += _gdk_offset_y;
-	    }
-  
-	  GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->width = client_rect.right - client_rect.left;
-	  GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl)->height = client_rect.bottom - client_rect.top;
-	  
-	  ((GdkWindowObject *) window)->x = point.x;
-	  ((GdkWindowObject *) window)->y = point.y;
-
 	  if (((GdkWindowObject *) window)->event_mask & GDK_STRUCTURE_MASK)
 	    {
-	      GdkEvent *event = gdk_event_new (GDK_CONFIGURE);
-	      
-	      event->configure.window = window;
-	      
-	      event->configure.width = client_rect.right - client_rect.left;
-	      event->configure.height = client_rect.bottom - client_rect.top;
-	      
-	      event->configure.x = point.x;
-	      event->configure.y = point.y;
-	      
 	      if (((GdkWindowObject *) window)->resize_count > 1)
 		((GdkWindowObject *) window)->resize_count -= 1;
 	      
-#if 0 /* I don't like calling _gdk_event_func() from here, isn't there
-       * a risk of getting events processed in the wrong order, like
-       * Owen says in the discussion of bug #99540?
-       */
-	      fixup_event (event);
-	      if (_gdk_event_func)
-		(*_gdk_event_func) (event, _gdk_event_data);
-	      gdk_event_free (event);
-#else /* Calling append_event() is slower, but guarantees that events won't
-       * get reordered, I think.
-       */
-	      append_event (event);
-#endif
-	      
+	      handle_configure_event (msg, window);
+	      g_main_context_iteration (NULL, FALSE);
+
 	      /* Dispatch main loop - to realize resizes... */
 	      handle_stuff_while_moving_or_resizing ();
 	      
@@ -3138,6 +2919,7 @@ gdk_event_translate (MSG  *msg,
 
     case WM_SIZING:
       GetWindowRect (GDK_WINDOW_HWND (window), &rect);
+      drag = (RECT *) msg->lParam;
       GDK_NOTE (EVENTS, g_print (" %s curr:%s drag:%s",
 				 (msg->wParam == WMSZ_BOTTOM ? "BOTTOM" :
 				  (msg->wParam == WMSZ_BOTTOMLEFT ? "BOTTOMLEFT" :
@@ -3150,13 +2932,13 @@ gdk_event_translate (MSG  *msg,
 					(msg->wParam == WMSZ_BOTTOMRIGHT ? "BOTTOMRIGHT" :
 					 "???")))))))),
 				 _gdk_win32_rect_to_string (&rect),
-				 _gdk_win32_rect_to_string ((RECT *) msg->lParam)));
+				 _gdk_win32_rect_to_string (drag)));
 
       impl = GDK_WINDOW_IMPL_WIN32 (((GdkWindowObject *) window)->impl);
-      drag = (RECT *) msg->lParam;
       orig_drag = *drag;
       if (impl->hint_flags & GDK_HINT_RESIZE_INC)
 	{
+	  GDK_NOTE (EVENTS, g_print (" (RESIZE_INC)"));
 	  if (impl->hint_flags & GDK_HINT_BASE_SIZE)
 	    {
 	      /* Resize in increments relative to the base size */
@@ -3185,7 +2967,6 @@ gdk_event_translate (MSG  *msg,
 	      if (drag->bottom == rect.bottom)
 		break;
 	      adjust_drag (&drag->bottom, rect.bottom, impl->hints.height_inc);
-
 	      break;
 
 	    case WMSZ_BOTTOMLEFT:
@@ -3240,7 +3021,7 @@ gdk_event_translate (MSG  *msg,
 	    {
 	      *ret_valp = TRUE;
 	      return_val = TRUE;
-	      GDK_NOTE (EVENTS, g_print (" (handled RESIZE_INC: drag:%s)",
+	      GDK_NOTE (EVENTS, g_print (" (handled RESIZE_INC: %s)",
 					 _gdk_win32_rect_to_string (drag)));
 	    }
 	}
@@ -3249,18 +3030,105 @@ gdk_event_translate (MSG  *msg,
 
       if (impl->hint_flags & GDK_HINT_ASPECT)
 	{
-	  gdouble drag_aspect = (gdouble) (drag->right - drag->left) / (drag->bottom - drag->top);
-	  
-	  GDK_NOTE (EVENTS, g_print (" (aspect:%g)", drag_aspect));
-	  if (drag_aspect < impl->hints.min_aspect ||
-	      drag_aspect > impl->hints.max_aspect)
+	  RECT decorated_rect;
+	  RECT undecorated_drag;
+	  int decoration_width, decoration_height;
+	  gdouble drag_aspect;
+	  int drag_width, drag_height, new_width, new_height;
+
+	  GetClientRect (GDK_WINDOW_HWND (window), &rect);
+	  decorated_rect = rect;
+	  _gdk_win32_adjust_client_rect (window, &decorated_rect);
+
+	  /* Set undecorated_drag to the client area being dragged
+	   * out, in screen coordinates.
+	   */
+	  undecorated_drag = *drag;
+	  undecorated_drag.left -= decorated_rect.left - rect.left;
+	  undecorated_drag.right -= decorated_rect.right - rect.right;
+	  undecorated_drag.top -= decorated_rect.top - rect.top;
+	  undecorated_drag.bottom -= decorated_rect.bottom - rect.bottom;
+
+	  decoration_width = (decorated_rect.right - decorated_rect.left) - (rect.right - rect.left);
+	  decoration_height = (decorated_rect.bottom - decorated_rect.top) - (rect.bottom - rect.top);
+
+	  drag_width = undecorated_drag.right - undecorated_drag.left;
+	  drag_height = undecorated_drag.bottom - undecorated_drag.top;
+
+	  drag_aspect = (gdouble) drag_width / drag_height;
+
+	  GDK_NOTE (EVENTS, g_print (" (ASPECT:%g--%g curr: %g)",
+				     impl->hints.min_aspect, impl->hints.max_aspect, drag_aspect));
+
+	  if (drag_aspect < impl->hints.min_aspect)
 	    {
-	      *drag = rect;
-	      *ret_valp = TRUE;
-	      return_val = TRUE;
-	      GDK_NOTE (EVENTS, g_print (" (handled ASPECT: drag:%s)",
-					 _gdk_win32_rect_to_string (drag)));
+	      /* Aspect is getting too narrow */
+	      switch (msg->wParam)
+		{
+		case WMSZ_BOTTOM:
+		case WMSZ_TOP:
+		  /* User drags top or bottom edge outward. Keep height, increase width. */
+		  new_width = impl->hints.min_aspect * drag_height;
+		  drag->left -= (new_width - drag_width) / 2;
+		  drag->right = drag->left + new_width + decoration_width;
+		  break;
+		case WMSZ_BOTTOMLEFT:
+		case WMSZ_BOTTOMRIGHT:
+		  /* User drags bottom-left or bottom-right corner down. Adjust height. */
+		  new_height = drag_width / impl->hints.min_aspect;
+		  drag->bottom = drag->top + new_height + decoration_height;
+		  break;
+		case WMSZ_LEFT:
+		case WMSZ_RIGHT:
+		  /* User drags left or right edge inward. Decrease height */
+		  new_height = drag_width / impl->hints.min_aspect;
+		  drag->top += (drag_height - new_height) / 2;
+		  drag->bottom = drag->top + new_height + decoration_height;
+		  break;
+		case WMSZ_TOPLEFT:
+		case WMSZ_TOPRIGHT:
+		  /* User drags top-left or top-right corner up. Adjust height. */
+		  new_height = drag_width / impl->hints.min_aspect;
+		  drag->top = drag->bottom - new_height - decoration_height;
+		}
 	    }
+	  else if (drag_aspect > impl->hints.max_aspect)
+	    {
+	      /* Aspect is getting too wide */
+	      switch (msg->wParam)
+		{
+		case WMSZ_BOTTOM:
+		case WMSZ_TOP:
+		  /* User drags top or bottom edge inward. Decrease width. */
+		  new_width = impl->hints.max_aspect * drag_height;
+		  drag->left += (drag_width - new_width) / 2;
+		  drag->right = drag->left + new_width + decoration_width;
+		  break;
+		case WMSZ_BOTTOMLEFT:
+		case WMSZ_TOPLEFT:
+		  /* User drags bottom-left or top-left corner left. Adjust width. */
+		  new_width = impl->hints.max_aspect * drag_height;
+		  drag->left = drag->right - new_width - decoration_width;
+		  break;
+		case WMSZ_BOTTOMRIGHT:
+		case WMSZ_TOPRIGHT:
+		  /* User drags bottom-right or top-right corner right. Adjust width. */
+		  new_width = impl->hints.max_aspect * drag_height;
+		  drag->right = drag->left + new_width + decoration_width;
+		  break;
+		case WMSZ_LEFT:
+		case WMSZ_RIGHT:
+		  /* User drags left or right edge outward. Increase height. */
+		  new_height = drag_width / impl->hints.max_aspect;
+		  drag->top -= (new_height - drag_height) / 2;
+		  drag->bottom = drag->top + new_height + decoration_height;
+		  break;
+		}
+	    }
+	  *ret_valp = TRUE;
+	  return_val = TRUE;
+	  GDK_NOTE (EVENTS, g_print (" (handled ASPECT: %s)",
+				     _gdk_win32_rect_to_string (drag)));
 	}
       break;
 
@@ -3308,7 +3176,7 @@ gdk_event_translate (MSG  *msg,
 
       if (impl->hint_flags & (GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE))
 	{
-	  /* Don't call DefWindowProc() */
+	  /* Don't call DefWindowProcW() */
 	  GDK_NOTE (EVENTS, g_print (" (handled, mintrack:%ldx%ld maxtrack:%ldx%ld "
 				     "maxpos:%+ld%+ld maxsize:%ldx%ld)",
 				     mmi->ptMinTrackSize.x, mmi->ptMinTrackSize.y,
@@ -3441,7 +3309,6 @@ gdk_event_translate (MSG  *msg,
 	}
       break;
 
-#ifdef HAVE_WINTAB
     case WM_ACTIVATE:
       /* Bring any tablet contexts to the top of the overlap order when
        * one of our windows is activated.
@@ -3482,7 +3349,6 @@ gdk_event_translate (MSG  *msg,
       else
 	gdk_event_free (event);
       break;
-#endif
     }
 
 done:
@@ -3499,11 +3365,14 @@ _gdk_events_queue (GdkDisplay *display)
 {
   MSG msg;
 
+  if (modal_win32_dialog != NULL)
+    return;
+  
   while (!_gdk_event_queue_find_first (display) &&
-	 PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
+	 PeekMessageW (&msg, NULL, 0, 0, PM_REMOVE))
     {
       TranslateMessage (&msg);
-      DispatchMessage (&msg);
+      DispatchMessageW (&msg);
     }
 }
 
@@ -3519,7 +3388,8 @@ gdk_event_prepare (GSource *source,
   *timeout = -1;
 
   retval = (_gdk_event_queue_find_first (_gdk_display) != NULL ||
-	    PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE));
+	    (modal_win32_dialog == NULL &&
+	     PeekMessageW (&msg, NULL, 0, 0, PM_NOREMOVE)));
 
   GDK_THREADS_LEAVE ();
 
@@ -3536,7 +3406,8 @@ gdk_event_check (GSource *source)
 
   if (event_poll_fd.revents & G_IO_IN)
     retval = (_gdk_event_queue_find_first (_gdk_display) != NULL ||
-	      PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE));
+	      (modal_win32_dialog == NULL &&
+	       PeekMessageW (&msg, NULL, 0, 0, PM_NOREMOVE)));
   else
     retval = FALSE;
 
@@ -3570,6 +3441,12 @@ gdk_event_dispatch (GSource     *source,
   return TRUE;
 }
 
+void
+gdk_win32_set_modal_dialog_libgtk_only (HWND window)
+{
+  modal_win32_dialog = window;
+}
+
 static void
 check_for_too_much_data (GdkEvent *event)
 {
@@ -3587,9 +3464,9 @@ gdk_event_send_client_message_for_display (GdkDisplay     *display,
 {
   check_for_too_much_data (event);
 
-  return PostMessage ((HWND) winid, client_message,
-		      (WPARAM) event->client.message_type,
-		      event->client.data.l[0]);
+  return PostMessageW ((HWND) winid, client_message,
+		       (WPARAM) event->client.message_type,
+		       event->client.data.l[0]);
 }
 
 void
@@ -3598,9 +3475,9 @@ gdk_screen_broadcast_client_message (GdkScreen *screen,
 {
   check_for_too_much_data (event);
 
-  PostMessage (HWND_BROADCAST, client_message,
+  PostMessageW (HWND_BROADCAST, client_message,
 	       (WPARAM) event->client.message_type,
-	       event->client.data.l[0]);
+		event->client.data.l[0]);
 }
 
 void
@@ -3610,10 +3487,10 @@ gdk_flush (void)
   MSG msg;
 
   /* Process all messages currently available */
-  while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
+  while (PeekMessageW (&msg, NULL, 0, 0, PM_REMOVE))
     {
       TranslateMessage (&msg);
-      DispatchMessage (&msg);
+      DispatchMessageW (&msg);
     }
 #endif
 
@@ -3628,8 +3505,8 @@ gdk_display_sync (GdkDisplay * display)
   g_return_if_fail (display == _gdk_display);
 
   /* Process all messages currently available */
-  while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE))
-    DispatchMessage (&msg);
+  while (PeekMessageW (&msg, NULL, 0, 0, PM_REMOVE))
+    DispatchMessageW (&msg);
 }
 
 void
