@@ -25,8 +25,11 @@
 
 #include <config.h>
 
+#include <atk/atk.h>
+
 #include "gtkassistant.h"
 
+#include "gtkaccessible.h"
 #include "gtkbutton.h"
 #include "gtkhbox.h"
 #include "gtkhbbox.h"
@@ -114,6 +117,8 @@ static void     gtk_assistant_get_child_property (GtkContainer      *container,
 						  GValue            *value,
 						  GParamSpec        *pspec);
 
+static AtkObject *gtk_assistant_get_accessible   (GtkWidget         *widget);
+
 enum
 {
   CHILD_PROP_0,
@@ -136,7 +141,7 @@ enum
 static guint signals [LAST_SIGNAL] = { 0 };
 
 
-G_DEFINE_TYPE (GtkAssistant, gtk_assistant, GTK_TYPE_WINDOW);
+G_DEFINE_TYPE (GtkAssistant, gtk_assistant, GTK_TYPE_WINDOW)
 
 
 static void
@@ -162,6 +167,7 @@ gtk_assistant_class_init (GtkAssistantClass *class)
   widget_class->delete_event = gtk_assistant_delete_event;
   widget_class->expose_event = gtk_assistant_expose;
   widget_class->focus = gtk_assistant_focus;
+  widget_class->get_accessible = gtk_assistant_get_accessible;
 
   container_class->add = gtk_assistant_add;
   container_class->remove = gtk_assistant_remove;
@@ -217,7 +223,7 @@ gtk_assistant_class_init (GtkAssistantClass *class)
    * A handler for the ::apply signal should carry out the actions for which the
    * wizard has collected data. If the action takes a long time to complete, you
    * might consider to put a page of type GTK_ASSISTANT_PAGE_PROGRESS after the
-   * confirmation page and handle this operation within the "prepare" signal of
+   * confirmation page and handle this operation within the ::prepare signal of
    * the progress page.
    *
    * Since: 2.10
@@ -672,6 +678,8 @@ gtk_assistant_init (GtkAssistant *assistant)
 
   priv = assistant->priv = GTK_ASSISTANT_GET_PRIVATE (assistant);
 
+  gtk_container_set_reallocate_redraws (GTK_CONTAINER (assistant), TRUE);
+
   gtk_widget_push_composite_child ();
 
   /* Header */
@@ -841,12 +849,29 @@ remove_page (GtkAssistant *assistant,
 {
   GtkAssistantPrivate *priv = assistant->priv;
   GtkAssistantPage *page_info;
+  GList *page_node;
 
   page_info = element->data;
 
-  /* If we are mapped and visible, we want to deal with changing the page. */
-  if ((GTK_WIDGET_MAPPED (page_info->page)) && (page_info == priv->current_page))
-    compute_next_step (assistant);
+  /* If this is the current page, we need to switch away. */
+  if (page_info == priv->current_page)
+    {
+      if (!compute_next_step (assistant))
+        {
+  	  /* The best we can do at this point is probably to pick the first
+	   * visible page.
+	   */
+	  page_node = priv->pages;
+
+	  while (page_node && !GTK_WIDGET_VISIBLE (((GtkAssistantPage *) page_node->data)->page))
+	    page_node = page_node->next;
+
+	  if (page_node)
+	    priv->current_page = page_node->data;
+	  else
+	    priv->current_page = NULL;
+        }
+    }
 
   priv->pages = g_list_remove_link (priv->pages, element);
   priv->visited_pages = g_slist_remove_all (priv->visited_pages, page_info);
@@ -1319,7 +1344,8 @@ gtk_assistant_focus (GtkWidget        *widget,
   if (container->focus_child == priv->action_area)
     {
       if (!gtk_widget_child_focus (priv->action_area, direction) &&
-	  !gtk_widget_child_focus (priv->current_page->page, direction))
+	  (priv->current_page == NULL ||
+	   !gtk_widget_child_focus (priv->current_page->page, direction)))
 	{
 	  /* if we're leaving the action area and the current page hasn't
 	     any focusable widget, clear focus and go back to the action area */
@@ -1329,13 +1355,15 @@ gtk_assistant_focus (GtkWidget        *widget,
     }
   else
     {
-      if (!gtk_widget_child_focus (priv->current_page->page, direction) &&
+      if ((priv->current_page ==  NULL ||
+	   !gtk_widget_child_focus (priv->current_page->page, direction)) &&
 	  !gtk_widget_child_focus (priv->action_area, direction))
 	{
 	  /* if we're leaving the current page and there isn't nothing focusable
 	     in the action area, try to clear focus and go back to the page */
 	  gtk_window_set_focus (GTK_WINDOW (widget), NULL);
-	  gtk_widget_child_focus (priv->current_page->page, direction);
+	  if (priv->current_page != NULL)
+	    gtk_widget_child_focus (priv->current_page->page, direction);
 	}
     }
 
@@ -1758,7 +1786,7 @@ gtk_assistant_remove_action_widget (GtkAssistant *assistant,
 /**
  * gtk_assistant_set_page_title:
  * @assistant: a #GtkAssistant
- * @page: a page of @assitant
+ * @page: a page of @assistant
  * @title: the new title for @page
  * 
  * Sets a title for @page. The title is displayed in the header
@@ -1771,14 +1799,12 @@ gtk_assistant_set_page_title (GtkAssistant *assistant,
 			      GtkWidget    *page,
 			      const gchar  *title)
 {
-  GtkAssistantPrivate *priv;
   GtkAssistantPage *page_info;
   GList *child;
 
   g_return_if_fail (GTK_IS_ASSISTANT (assistant));
   g_return_if_fail (GTK_IS_WIDGET (page));
 
-  priv = assistant->priv;
   child = find_page (assistant, page);
 
   g_return_if_fail (child != NULL);
@@ -1805,14 +1831,12 @@ G_CONST_RETURN gchar*
 gtk_assistant_get_page_title (GtkAssistant *assistant,
 			      GtkWidget    *page)
 {
-  GtkAssistantPrivate *priv;
   GtkAssistantPage *page_info;
   GList *child;
 
   g_return_val_if_fail (GTK_IS_ASSISTANT (assistant), NULL);
   g_return_val_if_fail (GTK_IS_WIDGET (page), NULL);
 
-  priv = assistant->priv;
   child = find_page (assistant, page);
 
   g_return_val_if_fail (child != NULL, NULL);
@@ -1825,7 +1849,7 @@ gtk_assistant_get_page_title (GtkAssistant *assistant,
 /**
  * gtk_assistant_set_page_type:
  * @assistant: a #GtkAssistant
- * @page: a page of @assitant
+ * @page: a page of @assistant
  * @type: the new type for @page
  * 
  * Sets the page type for @page. The page type determines the page
@@ -1880,14 +1904,12 @@ GtkAssistantPageType
 gtk_assistant_get_page_type (GtkAssistant *assistant,
 			     GtkWidget    *page)
 {
-  GtkAssistantPrivate *priv;
   GtkAssistantPage *page_info;
   GList *child;
 
   g_return_val_if_fail (GTK_IS_ASSISTANT (assistant), GTK_ASSISTANT_PAGE_CONTENT);
   g_return_val_if_fail (GTK_IS_WIDGET (page), GTK_ASSISTANT_PAGE_CONTENT);
 
-  priv = assistant->priv;
   child = find_page (assistant, page);
 
   g_return_val_if_fail (child != NULL, GTK_ASSISTANT_PAGE_CONTENT);
@@ -1900,7 +1922,7 @@ gtk_assistant_get_page_type (GtkAssistant *assistant,
 /**
  * gtk_assistant_set_page_header_image:
  * @assistant: a #GtkAssistant
- * @page: a page of @assitant
+ * @page: a page of @assistant
  * @pixbuf: the new header image @page
  * 
  * Sets a header image for @page. This image is displayed in the header
@@ -1962,14 +1984,12 @@ GdkPixbuf*
 gtk_assistant_get_page_header_image (GtkAssistant *assistant,
 				     GtkWidget    *page)
 {
-  GtkAssistantPrivate *priv;
   GtkAssistantPage *page_info;
   GList *child;
 
   g_return_val_if_fail (GTK_IS_ASSISTANT (assistant), NULL);
   g_return_val_if_fail (GTK_IS_WIDGET (page), NULL);
 
-  priv = assistant->priv;
   child = find_page (assistant, page);
 
   g_return_val_if_fail (child != NULL, NULL);
@@ -1982,7 +2002,7 @@ gtk_assistant_get_page_header_image (GtkAssistant *assistant,
 /**
  * gtk_assistant_set_page_side_image:
  * @assistant: a #GtkAssistant
- * @page: a page of @assitant
+ * @page: a page of @assistant
  * @pixbuf: the new header image @page
  * 
  * Sets a header image for @page. This image is displayed in the side
@@ -2044,14 +2064,12 @@ GdkPixbuf*
 gtk_assistant_get_page_side_image (GtkAssistant *assistant,
 				   GtkWidget    *page)
 {
-  GtkAssistantPrivate *priv;
   GtkAssistantPage *page_info;
   GList *child;
 
   g_return_val_if_fail (GTK_IS_ASSISTANT (assistant), NULL);
   g_return_val_if_fail (GTK_IS_WIDGET (page), NULL);
 
-  priv = assistant->priv;
   child = find_page (assistant, page);
 
   g_return_val_if_fail (child != NULL, NULL);
@@ -2064,7 +2082,7 @@ gtk_assistant_get_page_side_image (GtkAssistant *assistant,
 /**
  * gtk_assistant_set_page_complete:
  * @assistant: a #GtkAssistant
- * @page: a page of @assitant
+ * @page: a page of @assistant
  * @complete: the completeness status of the page
  * 
  * Sets whether @page contents are complete. This will make
@@ -2119,14 +2137,12 @@ gboolean
 gtk_assistant_get_page_complete (GtkAssistant *assistant,
 				 GtkWidget    *page)
 {
-  GtkAssistantPrivate *priv;
   GtkAssistantPage *page_info;
   GList *child;
 
   g_return_val_if_fail (GTK_IS_ASSISTANT (assistant), FALSE);
   g_return_val_if_fail (GTK_IS_WIDGET (page), FALSE);
 
-  priv = assistant->priv;
   child = find_page (assistant, page);
 
   g_return_val_if_fail (child != NULL, FALSE);
@@ -2161,5 +2177,189 @@ gtk_assistant_update_buttons_state (GtkAssistant *assistant)
 }
 
 
+
+/* accessible implementation */
+
+static gint
+gtk_assistant_accessible_get_n_children (AtkObject *accessible)
+{
+  GtkAssistant *assistant;
+  GtkWidget *widget;
+
+  widget = GTK_ACCESSIBLE (accessible)->widget;
+
+  if (!widget)
+    return 0;
+
+  assistant = GTK_ASSISTANT (widget);
+   
+  return g_list_length (assistant->priv->pages) + 1;
+}
+
+
+static AtkObject *
+gtk_assistant_accessible_ref_child (AtkObject *accessible,
+				    gint       index)
+{
+  GtkAssistant *assistant;
+  GtkAssistantPrivate *priv;
+  GtkWidget *widget, *child;
+  gint n_pages;
+  AtkObject *obj;
+
+  widget = GTK_ACCESSIBLE (accessible)->widget;
+  if (!widget)
+    return NULL;
+
+  assistant = GTK_ASSISTANT (widget);
+  priv = assistant->priv;
+  n_pages = g_list_length (priv->pages);
+
+  if (index < 0)
+    return NULL;
+  else if (index < n_pages)
+    {
+      GtkAssistantPage *page = g_list_nth_data (priv->pages, index / 2);
+
+      child = page->page;
+    }
+  else if (index == n_pages)
+    {
+      child = priv->action_area;
+    }
+  else
+    return NULL;
+  
+  obj = gtk_widget_get_accessible (child);
+
+  return g_object_ref (obj);
+}
+
+static void
+gtk_assistant_accessible_class_init (AtkObjectClass *class)
+{
+  class->get_n_children = gtk_assistant_accessible_get_n_children;
+  class->ref_child = gtk_assistant_accessible_ref_child;
+}
+
+static GType
+gtk_assistant_accessible_get_type (void)
+{
+  static GType type = 0;
+  
+  if (!type)
+    {
+      /*
+       * Figure out the size of the class and instance
+       * we are deriving from
+       */
+      AtkObjectFactory *factory;
+      GType derived_type;
+      GTypeQuery query;
+      GType derived_atk_type;
+      
+      derived_type = g_type_parent (GTK_TYPE_ASSISTANT);
+      factory = atk_registry_get_factory (atk_get_default_registry (),
+					  derived_type);
+      derived_atk_type = atk_object_factory_get_accessible_type (factory);
+      g_type_query (derived_atk_type, &query);
+      
+      type = g_type_register_static_simple (derived_atk_type, 
+					    I_("GtkAssistantAccessible"), 
+					    query.class_size,
+					    (GClassInitFunc) gtk_assistant_accessible_class_init,
+					    query.instance_size,
+					    NULL, 0);
+    }
+
+  return type;
+}
+
+static AtkObject *
+gtk_assistant_accessible_new (GObject *obj)
+{
+  AtkObject *accessible;
+
+  g_return_val_if_fail (GTK_IS_ASSISTANT (obj), NULL);
+
+  accessible = g_object_new (gtk_assistant_accessible_get_type (), NULL); 
+  atk_object_initialize (accessible, obj);
+  
+  return accessible;
+}
+
+static GType
+gtk_assistant_accessible_factory_get_accessible_type (void)
+{
+  return gtk_assistant_accessible_get_type ();
+}
+
+static AtkObject*
+gtk_assistant_accessible_factory_create_accessible (GObject *obj)
+{
+  return gtk_assistant_accessible_new (obj);
+}
+
+static void
+gtk_assistant_accessible_factory_class_init (AtkObjectFactoryClass *class)
+{
+  class->create_accessible = gtk_assistant_accessible_factory_create_accessible;
+  class->get_accessible_type = gtk_assistant_accessible_factory_get_accessible_type;
+}
+
+static GType
+gtk_assistant_accessible_factory_get_type (void)
+{
+  static GType type = 0;
+
+  if (!type) 
+    {
+      type = g_type_register_static_simple (ATK_TYPE_OBJECT_FACTORY, 
+					    I_("GtkAssistantAccessibleFactory"),
+					    sizeof (AtkObjectFactoryClass),
+					    (GClassInitFunc) gtk_assistant_accessible_factory_class_init,
+					    sizeof (AtkObjectFactory),
+					    NULL, 0);
+    }
+  
+  return type;
+}
+
+static AtkObject *
+gtk_assistant_get_accessible (GtkWidget *widget)
+{
+  static gboolean first_time = TRUE;
+
+  if (first_time) 
+    {
+      AtkObjectFactory *factory;
+      AtkRegistry *registry;
+      GType derived_type; 
+      GType derived_atk_type; 
+
+      /*
+       * Figure out whether accessibility is enabled by looking at the
+       * type of the accessible object which would be created for
+       * the parent type of GtkAssistant.
+       */
+      derived_type = g_type_parent (GTK_TYPE_ASSISTANT);
+
+      registry = atk_get_default_registry ();
+      factory = atk_registry_get_factory (registry,
+					  derived_type);
+      derived_atk_type = atk_object_factory_get_accessible_type (factory);
+      if (g_type_is_a (derived_atk_type, GTK_TYPE_ACCESSIBLE))
+	{
+	  atk_registry_set_factory_type (registry, 
+					 GTK_TYPE_ASSISTANT,
+					 gtk_assistant_accessible_factory_get_type ());
+	}
+      first_time = FALSE;
+    }
+
+  return GTK_WIDGET_CLASS (gtk_assistant_parent_class)->get_accessible (widget);
+}
+
+
 #define __GTK_ASSISTANT_C__
 #include "gtkaliasdef.c"
