@@ -24,6 +24,7 @@
 #include "gtksignal.h"
 #include "gtkcellrenderertext.h"
 #include "gtkcellrendererpixbuf.h"
+#include "gtkprivate.h"
 #include <gobject/gmarshal.h>
 #include "gtkalias.h"
 
@@ -56,7 +57,6 @@ struct _GtkCellViewPrivate
 };
 
 
-static void        gtk_cell_view_class_init               (GtkCellViewClass *klass);
 static void        gtk_cell_view_cell_layout_init         (GtkCellLayoutIface *iface);
 static void        gtk_cell_view_get_property             (GObject           *object,
                                                            guint             param_id,
@@ -66,10 +66,7 @@ static void        gtk_cell_view_set_property             (GObject          *obj
                                                            guint             param_id,
                                                            const GValue     *value,
                                                            GParamSpec       *pspec);
-static void        gtk_cell_view_init                     (GtkCellView      *cellview);
 static void        gtk_cell_view_finalize                 (GObject          *object);
-static void        gtk_cell_view_style_set                (GtkWidget        *widget,
-                                                           GtkStyle         *previous_style);
 static void        gtk_cell_view_size_request             (GtkWidget        *widget,
                                                            GtkRequisition   *requisition);
 static void        gtk_cell_view_size_allocate            (GtkWidget        *widget,
@@ -106,7 +103,7 @@ static void       gtk_cell_view_cell_layout_set_cell_data_func (GtkCellLayout
 static void       gtk_cell_view_cell_layout_reorder            (GtkCellLayout         *layout,
                                                                 GtkCellRenderer       *cell,
                                                                 gint                   position);
-
+static GList *    gtk_cell_view_cell_layout_get_cells          (GtkCellLayout         *layout);
 
 #define GTK_CELL_VIEW_GET_PRIVATE(obj)    (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GTK_TYPE_CELL_VIEW, GtkCellViewPrivate))
 
@@ -115,48 +112,13 @@ enum
   PROP_0,
   PROP_BACKGROUND,
   PROP_BACKGROUND_GDK,
-  PROP_BACKGROUND_SET
+  PROP_BACKGROUND_SET,
+  PROP_MODEL
 };
 
-static GtkObjectClass *parent_class = NULL;
-
-
-GType
-gtk_cell_view_get_type (void)
-{
-  static GType cell_view_type = 0;
-
-  if (!cell_view_type)
-    {
-      static const GTypeInfo cell_view_info =
-        {
-          sizeof (GtkCellViewClass),
-          NULL, /* base_init */
-          NULL, /* base_finalize */
-          (GClassInitFunc) gtk_cell_view_class_init,
-          NULL, /* class_finalize */
-          NULL, /* class_data */
-          sizeof (GtkCellView),
-          0,
-          (GInstanceInitFunc) gtk_cell_view_init
-        };
-
-      static const GInterfaceInfo cell_layout_info =
-       {
-         (GInterfaceInitFunc) gtk_cell_view_cell_layout_init,
-         NULL,
-         NULL
-       };
-
-      cell_view_type = g_type_register_static (GTK_TYPE_WIDGET, "GtkCellView",
-                                               &cell_view_info, 0);
-
-      g_type_add_interface_static (cell_view_type, GTK_TYPE_CELL_LAYOUT,
-                                   &cell_layout_info);
-    }
-
-  return cell_view_type;
-}
+G_DEFINE_TYPE_WITH_CODE (GtkCellView, gtk_cell_view, GTK_TYPE_WIDGET, 
+			 G_IMPLEMENT_INTERFACE (GTK_TYPE_CELL_LAYOUT,
+						gtk_cell_view_cell_layout_init))
 
 static void
 gtk_cell_view_class_init (GtkCellViewClass *klass)
@@ -164,8 +126,6 @@ gtk_cell_view_class_init (GtkCellViewClass *klass)
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
 
-  parent_class = g_type_class_peek_parent (klass);
-
   gobject_class->get_property = gtk_cell_view_get_property;
   gobject_class->set_property = gtk_cell_view_set_property;
   gobject_class->finalize = gtk_cell_view_finalize;
@@ -173,7 +133,6 @@ gtk_cell_view_class_init (GtkCellViewClass *klass)
   widget_class->expose_event = gtk_cell_view_expose;
   widget_class->size_allocate = gtk_cell_view_size_allocate;
   widget_class->size_request = gtk_cell_view_size_request;
-  widget_class->style_set = gtk_cell_view_style_set;
 
   /* properties */
   g_object_class_install_property (gobject_class,
@@ -182,18 +141,33 @@ gtk_cell_view_class_init (GtkCellViewClass *klass)
                                                         P_("Background color name"),
                                                         P_("Background color as a string"),
                                                         NULL,
-                                                        G_PARAM_WRITABLE));
+                                                        GTK_PARAM_WRITABLE));
   g_object_class_install_property (gobject_class,
                                    PROP_BACKGROUND_GDK,
-                                   g_param_spec_boxed ("background_gdk",
+                                   g_param_spec_boxed ("background-gdk",
                                                       P_("Background color"),
                                                       P_("Background color as a GdkColor"),
                                                       GDK_TYPE_COLOR,
-                                                      G_PARAM_READABLE | G_PARAM_WRITABLE));
+                                                      GTK_PARAM_READWRITE));
 
-#define ADD_SET_PROP(propname, propval, nick, blurb) g_object_class_install_property (gobject_class, propval, g_param_spec_boolean (propname, nick, blurb, FALSE, G_PARAM_READABLE | G_PARAM_WRITABLE))
+  /**
+   * GtkCellView:model
+   *
+   * The model for cell view
+   *
+   * since 2.10
+   */
+  g_object_class_install_property (gobject_class,
+				   PROP_MODEL,
+				   g_param_spec_object  ("model",
+							 P_("CellView model"),
+							 P_("The model for cell view"),
+							 GTK_TYPE_TREE_MODEL,
+							 GTK_PARAM_READWRITE));
+  
+#define ADD_SET_PROP(propname, propval, nick, blurb) g_object_class_install_property (gobject_class, propval, g_param_spec_boolean (propname, nick, blurb, FALSE, GTK_PARAM_READWRITE))
 
-  ADD_SET_PROP ("background_set", PROP_BACKGROUND_SET,
+  ADD_SET_PROP ("background-set", PROP_BACKGROUND_SET,
                 P_("Background set"),
                 P_("Whether this tag affects the background color"));
 
@@ -210,6 +184,7 @@ gtk_cell_view_cell_layout_init (GtkCellLayoutIface *iface)
   iface->set_cell_data_func = gtk_cell_view_cell_layout_set_cell_data_func;
   iface->clear_attributes = gtk_cell_view_cell_layout_clear_attributes;
   iface->reorder = gtk_cell_view_cell_layout_reorder;
+  iface->get_cells = gtk_cell_view_cell_layout_get_cells;
 }
 
 static void
@@ -234,6 +209,9 @@ gtk_cell_view_get_property (GObject    *object,
       case PROP_BACKGROUND_SET:
         g_value_set_boolean (value, view->priv->background_set);
         break;
+      case PROP_MODEL:
+	g_value_set_object (value, view->priv->model);
+	break;
       default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
         break;
@@ -270,7 +248,10 @@ gtk_cell_view_set_property (GObject      *object,
       case PROP_BACKGROUND_SET:
         view->priv->background_set = g_value_get_boolean (value);
         break;
-      default:
+      case PROP_MODEL:
+	gtk_cell_view_set_model (view, g_value_get_object (value));
+	break;
+    default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
         break;
     }
@@ -284,15 +265,6 @@ gtk_cell_view_init (GtkCellView *cellview)
   cellview->priv = GTK_CELL_VIEW_GET_PRIVATE (cellview);
 }
 
-static void
-gtk_cell_view_style_set (GtkWidget *widget,
-                         GtkStyle  *previous_style)
-{
-  if (previous_style && GTK_WIDGET_REALIZED (widget))
-    gdk_window_set_background (widget->window,
-                               &widget->style->base[GTK_WIDGET_STATE (widget)]);
-}
-
 static void
 gtk_cell_view_finalize (GObject *object)
 {
@@ -306,7 +278,7 @@ gtk_cell_view_finalize (GObject *object)
   if (cellview->priv->displayed_row)
      gtk_tree_row_reference_free (cellview->priv->displayed_row);
 
-  (* G_OBJECT_CLASS (parent_class)->finalize) (object);
+  (* G_OBJECT_CLASS (gtk_cell_view_parent_class)->finalize) (object);
 }
 
 static void
@@ -428,23 +400,16 @@ gtk_cell_view_expose (GtkWidget      *widget,
   /* "blank" background */
   if (cellview->priv->background_set)
     {
-      GdkGC *gc;
+      cairo_t *cr = gdk_cairo_create (GTK_WIDGET (cellview)->window);
 
-      gc = gdk_gc_new (GTK_WIDGET (cellview)->window);
-      gdk_gc_set_rgb_fg_color (gc, &cellview->priv->background);
+      gdk_cairo_rectangle (cr, &widget->allocation);
+      cairo_set_source_rgb (cr,
+			    cellview->priv->background.red / 65535.,
+			    cellview->priv->background.green / 65535.,
+			    cellview->priv->background.blue / 65535.);
+      cairo_fill (cr);
 
-      gdk_draw_rectangle (GTK_WIDGET (cellview)->window,
-                          gc,
-                          TRUE,
-
-                          /*0, 0,*/
-                          widget->allocation.x,
-                          widget->allocation.y,
-
-                          widget->allocation.width,
-                          widget->allocation.height);
-
-      g_object_unref (gc);
+      cairo_destroy (cr);
     }
 
   /* set cell data (if available) */
@@ -596,8 +561,7 @@ gtk_cell_view_cell_layout_pack_start (GtkCellLayout   *layout,
   g_return_if_fail (GTK_IS_CELL_RENDERER (renderer));
   g_return_if_fail (!gtk_cell_view_get_cell_info (cellview, renderer));
 
-  g_object_ref (renderer);
-  gtk_object_sink (GTK_OBJECT (renderer));
+  g_object_ref_sink (renderer);
 
   info = g_new0 (GtkCellViewCellInfo, 1);
   info->cell = renderer;
@@ -619,8 +583,7 @@ gtk_cell_view_cell_layout_pack_end (GtkCellLayout   *layout,
   g_return_if_fail (GTK_IS_CELL_RENDERER (renderer));
   g_return_if_fail (!gtk_cell_view_get_cell_info (cellview, renderer));
 
-  g_object_ref (renderer);
-  gtk_object_sink (GTK_OBJECT (renderer));
+  g_object_ref_sink (renderer);
 
   info = g_new0 (GtkCellViewCellInfo, 1);
   info->cell = renderer;
@@ -912,7 +875,7 @@ gtk_cell_view_set_model (GtkCellView  *cell_view,
                          GtkTreeModel *model)
 {
   g_return_if_fail (GTK_IS_CELL_VIEW (cell_view));
-  g_return_if_fail (GTK_IS_TREE_MODEL (model));
+  g_return_if_fail (model == NULL || GTK_IS_TREE_MODEL (model));
 
   if (cell_view->priv->model)
     {
@@ -1098,5 +1061,12 @@ gtk_cell_view_get_cell_renderers (GtkCellView *cell_view)
   return g_list_reverse (retval);
 }
 
+static GList *
+gtk_cell_view_cell_layout_get_cells (GtkCellLayout *layout)
+{
+  return gtk_cell_view_get_cell_renderers (GTK_CELL_VIEW (layout));
+}
+
+
 #define __GTK_CELL_VIEW_C__
 #include "gtkaliasdef.c"
