@@ -73,6 +73,7 @@ struct _GtkIconThemePrivate
   guint pixbuf_supports_svg : 1;
   guint themes_valid        : 1;
   guint check_reload        : 1;
+  guint loading_themes      : 1;
   
   char *current_theme;
   char *fallback_theme;
@@ -199,13 +200,16 @@ static GtkIconInfo *theme_lookup_icon (IconTheme        *theme,
 static void         theme_list_icons  (IconTheme        *theme,
 				       GHashTable       *icons,
 				       GQuark            context);
+static void         theme_list_contexts  (IconTheme        *theme,
+					  GHashTable       *contexts);
 static void         theme_subdir_load (GtkIconTheme     *icon_theme,
 				       IconTheme        *theme,
 				       GKeyFile         *theme_file,
 				       char             *subdir);
 static void         do_theme_change   (GtkIconTheme     *icon_theme);
 
-static void  blow_themes               (GtkIconTheme    *icon_themes);
+static void     blow_themes               (GtkIconTheme    *icon_themes);
+static gboolean rescan_themes             (GtkIconTheme    *icon_themes);
 
 static void  icon_data_free            (GtkIconData     *icon_data);
 static void load_icon_data             (IconThemeDir    *dir,
@@ -377,6 +381,9 @@ display_closed (GdkDisplay   *display,
 static void
 update_current_theme (GtkIconTheme *icon_theme)
 {
+#define theme_changed(_old, _new) \
+  ((_old && !_new) || (!_old && _new) || \
+   (_old && _new && strcmp (_old, _new) != 0))
   GtkIconThemePrivate *priv = icon_theme->priv;
 
   if (!priv->custom_theme)
@@ -393,27 +400,25 @@ update_current_theme (GtkIconTheme *icon_theme)
 			"gtk-fallback-icon-theme", &fallback_theme, NULL);
 	}
 
-      if (!theme)
+      /* ensure that the current theme (even when just the default)
+       * is searched before any fallback theme
+       */
+      if (!theme && fallback_theme)
 	theme = g_strdup (DEFAULT_THEME_NAME);
 
-      if (strcmp (priv->current_theme, theme) != 0)
+      if (theme_changed (priv->current_theme, theme))
 	{
 	  g_free (priv->current_theme);
 	  priv->current_theme = theme;
-
 	  changed = TRUE;
 	}
       else
 	g_free (theme);
 
-      if ((priv->fallback_theme && !fallback_theme) ||
-	  (!priv->fallback_theme && fallback_theme) ||
-	  (priv->fallback_theme && fallback_theme &&
-	   strcmp (priv->fallback_theme, fallback_theme) != 0))
+      if (theme_changed (priv->fallback_theme, fallback_theme))
 	{
 	  g_free (priv->fallback_theme);
 	  priv->fallback_theme = fallback_theme;
-
 	  changed = TRUE;
 	}
       else
@@ -422,6 +427,7 @@ update_current_theme (GtkIconTheme *icon_theme)
       if (changed)
 	do_theme_change (icon_theme);
     }
+#undef theme_changed
 }
 
 /* Callback when the icon theme GtkSetting changes
@@ -507,13 +513,15 @@ gtk_icon_theme_set_screen (GtkIconTheme *icon_theme,
 static gboolean
 pixbuf_supports_svg (void)
 {
-  GSList *formats = gdk_pixbuf_get_formats ();
+  GSList *formats;
   GSList *tmp_list;
   static gint found_svg = -1;
 
   if (found_svg != -1)
     return found_svg;
- 
+
+  formats = gdk_pixbuf_get_formats ();
+
   found_svg = FALSE; 
   for (tmp_list = formats; tmp_list && !found_svg; tmp_list = tmp_list->next)
     {
@@ -546,7 +554,6 @@ gtk_icon_theme_init (GtkIconTheme *icon_theme)
   icon_theme->priv = priv;
 
   priv->custom_theme = FALSE;
-  priv->current_theme = g_strdup (DEFAULT_THEME_NAME);
 
   xdg_data_dirs = g_get_system_data_dirs ();
   for (i = 0; xdg_data_dirs[i]; i++) ;
@@ -589,8 +596,6 @@ reset_styles_idle (gpointer user_data)
   GtkIconTheme *icon_theme;
   GtkIconThemePrivate *priv;
 
-  GDK_THREADS_ENTER ();
-
   icon_theme = GTK_ICON_THEME (user_data);
   priv = icon_theme->priv;
 
@@ -602,8 +607,6 @@ reset_styles_idle (gpointer user_data)
 
   priv->reset_styles_idle = 0;
 
-  GDK_THREADS_LEAVE ();
-
   return FALSE;
 }
 
@@ -611,6 +614,9 @@ static void
 do_theme_change (GtkIconTheme *icon_theme)
 {
   GtkIconThemePrivate *priv = icon_theme->priv;
+
+  if (!priv->themes_valid)
+    return;
   
   GTK_NOTE (ICONTHEME, 
 	    g_print ("change to icon theme \"%s\"\n", priv->current_theme));
@@ -619,7 +625,7 @@ do_theme_change (GtkIconTheme *icon_theme)
 
   if (!priv->reset_styles_idle)
     priv->reset_styles_idle = 
-      g_idle_add_full (GTK_PRIORITY_RESIZE - 2, 
+      gdk_threads_add_idle_full (GTK_PRIORITY_RESIZE - 2, 
 		       reset_styles_idle, icon_theme, NULL);
 }
 
@@ -851,7 +857,7 @@ gtk_icon_theme_set_custom_theme (GtkIconTheme *icon_theme,
   if (theme_name != NULL)
     {
       priv->custom_theme = TRUE;
-      if (strcmp (theme_name, priv->current_theme) != 0)
+      if (!priv->current_theme || strcmp (theme_name, priv->current_theme) != 0)
 	{
 	  g_free (priv->current_theme);
 	  priv->current_theme = g_strdup (theme_name);
@@ -861,9 +867,12 @@ gtk_icon_theme_set_custom_theme (GtkIconTheme *icon_theme,
     }
   else
     {
-      priv->custom_theme = FALSE;
+      if (priv->custom_theme)
+	{
+	  priv->custom_theme = FALSE;
 
-      update_current_theme (icon_theme);
+	  update_current_theme (icon_theme);
+	}
     }
 }
 
@@ -875,7 +884,7 @@ insert_theme (GtkIconTheme *icon_theme, const char *theme_name)
   char **dirs;
   char **themes;
   GtkIconThemePrivate *priv;
-  IconTheme *theme;
+  IconTheme *theme = NULL;
   char *path;
   GKeyFile *theme_file;
   GError *error = NULL;
@@ -950,6 +959,8 @@ insert_theme (GtkIconTheme *icon_theme, const char *theme_name)
   if (!dirs)
     {
       g_warning ("Theme file for %s has no directories\n", theme_name);
+      priv->themes = g_list_remove (priv->themes, theme);
+      g_free (theme->name);
       g_free (theme->display_name);
       g_free (theme);
       g_key_file_free (theme_file);
@@ -992,10 +1003,8 @@ insert_theme (GtkIconTheme *icon_theme, const char *theme_name)
 static void
 free_unthemed_icon (UnthemedIcon *unthemed_icon)
 {
-  if (unthemed_icon->svg_filename)
-    g_free (unthemed_icon->svg_filename);
-  if (unthemed_icon->no_svg_filename)
-    g_free (unthemed_icon->no_svg_filename);
+  g_free (unthemed_icon->svg_filename);
+  g_free (unthemed_icon->no_svg_filename);
   g_slice_free (UnthemedIcon, unthemed_icon);
 }
 
@@ -1030,7 +1039,8 @@ load_themes (GtkIconTheme *icon_theme)
 
   priv->all_icons = g_hash_table_new (g_str_hash, g_str_equal);
   
-  insert_theme (icon_theme, priv->current_theme);
+  if (priv->current_theme)
+    insert_theme (icon_theme, priv->current_theme);
 
   /* Always look in the "default" icon theme, and in a fallback theme */
   if (priv->fallback_theme)
@@ -1047,20 +1057,21 @@ load_themes (GtkIconTheme *icon_theme)
       dir = icon_theme->priv->search_path[base];
 
       dir_mtime = g_slice_new (IconThemeDirMtime);
-      dir_mtime->cache = _gtk_icon_cache_new_for_path (dir);
-      dir_mtime->dir = g_strdup (dir);
-      if (g_stat (dir, &stat_buf) == 0 && S_ISDIR (stat_buf.st_mode))
-	dir_mtime->mtime = stat_buf.st_mtime;
-      else
-	dir_mtime->mtime = 0;
-      
       priv->dir_mtimes = g_list_append (priv->dir_mtimes, dir_mtime);
       
+      dir_mtime->dir = g_strdup (dir);
+      dir_mtime->mtime = 0;
+      dir_mtime->cache = NULL;
+
+      if (g_stat (dir, &stat_buf) != 0 || !S_ISDIR (stat_buf.st_mode))
+	continue;
+      dir_mtime->mtime = stat_buf.st_mtime;
+
+      dir_mtime->cache = _gtk_icon_cache_new_for_path (dir);
       if (dir_mtime->cache != NULL)
 	continue;
 
       gdir = g_dir_open (dir, 0, NULL);
-
       if (gdir == NULL)
 	continue;
 
@@ -1169,63 +1180,56 @@ ensure_valid_themes (GtkIconTheme *icon_theme)
   GTimeVal tv;
   gboolean was_valid = priv->themes_valid;
 
+  if (priv->loading_themes)
+    return;
+  priv->loading_themes = TRUE;
+
   _gtk_icon_theme_ensure_builtin_cache ();
 
   if (priv->themes_valid)
     {
       g_get_current_time (&tv);
 
-      if (ABS (tv.tv_sec - priv->last_stat_time) > 5)
-	gtk_icon_theme_rescan_if_needed (icon_theme);
+      if (ABS (tv.tv_sec - priv->last_stat_time) > 5 &&
+	  rescan_themes (icon_theme))
+	blow_themes (icon_theme);
     }
   
   if (!priv->themes_valid)
     {
       load_themes (icon_theme);
-      
-      if (!priv->check_reload && was_valid && priv->screen)
-	{	  
-	  static GdkAtom atom_iconthemes = GDK_NONE;
-	  GdkEvent *event = gdk_event_new (GDK_CLIENT_EVENT);
-	  int i;
 
-	  if (!atom_iconthemes)
-	    atom_iconthemes = gdk_atom_intern_static_string ("_GTK_LOAD_ICONTHEMES");
+      if (was_valid)
+	{
+	  g_signal_emit (icon_theme, signal_changed, 0);
 
-	  for (i = 0; i < 5; i++)
-	    event->client.data.l[i] = 0;
-	  event->client.data_format = 32;
-	  event->client.message_type = atom_iconthemes;
+	  if (!priv->check_reload && priv->screen)
+	    {	  
+	      static GdkAtom atom_iconthemes = GDK_NONE;
+	      GdkEvent *event = gdk_event_new (GDK_CLIENT_EVENT);
+	      int i;
 
-	  gdk_screen_broadcast_client_message (priv->screen, event);
+	      if (!atom_iconthemes)
+		atom_iconthemes = gdk_atom_intern_static_string ("_GTK_LOAD_ICONTHEMES");
+
+	      for (i = 0; i < 5; i++)
+		event->client.data.l[i] = 0;
+	      event->client.data_format = 32;
+	      event->client.message_type = atom_iconthemes;
+
+	      gdk_screen_broadcast_client_message (priv->screen, event);
+	    }
 	}
     }
+
+  priv->loading_themes = FALSE;
 }
 
-/**
- * gtk_icon_theme_lookup_icon:
- * @icon_theme: a #GtkIconTheme
- * @icon_name: the name of the icon to lookup
- * @size: desired icon size
- * @flags: flags modifying the behavior of the icon lookup
- * 
- * Looks up a named icon and returns a structure containing
- * information such as the filename of the icon. The icon
- * can then be rendered into a pixbuf using
- * gtk_icon_info_load_icon(). (gtk_icon_theme_load_icon()
- * combines these two steps if all you need is the pixbuf.)
- * 
- * Return value: a #GtkIconInfo structure containing information
- * about the icon, or %NULL if the icon wasn't found. Free with
- * gtk_icon_info_free()
- *
- * Since: 2.4
- **/
-GtkIconInfo *
-gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
-			    const gchar        *icon_name,
-			    gint                size,
-			    GtkIconLookupFlags  flags)
+static GtkIconInfo *
+choose_icon (GtkIconTheme       *icon_theme,
+	     const gchar        *icon_names[],
+	     gint                size,
+	     GtkIconLookupFlags  flags)
 {
   GtkIconThemePrivate *priv;
   GList *l;
@@ -1233,16 +1237,10 @@ gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
   UnthemedIcon *unthemed_icon;
   gboolean allow_svg;
   gboolean use_builtin;
-
-  g_return_val_if_fail (GTK_IS_ICON_THEME (icon_theme), NULL);
-  g_return_val_if_fail (icon_name != NULL, NULL);
-  g_return_val_if_fail ((flags & GTK_ICON_LOOKUP_NO_SVG) == 0 ||
-			(flags & GTK_ICON_LOOKUP_FORCE_SVG) == 0, NULL);
+  gint i;
 
   priv = icon_theme->priv;
 
-  GTK_NOTE (ICONTHEME, 
-	    g_print ("gtk_icon_theme_lookup_icon %s\n", icon_name));
   if (flags & GTK_ICON_LOOKUP_NO_SVG)
     allow_svg = FALSE;
   else if (flags & GTK_ICON_LOOKUP_FORCE_SVG)
@@ -1250,20 +1248,29 @@ gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
   else
     allow_svg = priv->pixbuf_supports_svg;
 
-  use_builtin = (flags & GTK_ICON_LOOKUP_USE_BUILTIN);
-
+  use_builtin = flags & GTK_ICON_LOOKUP_USE_BUILTIN;
+  
   ensure_valid_themes (icon_theme);
 
   for (l = priv->themes; l; l = l->next)
     {
       IconTheme *theme = l->data;
       
-      icon_info = theme_lookup_icon (theme, icon_name, size, allow_svg, use_builtin);
-      if (icon_info)
-	goto out;
+      for (i = 0; icon_names[i]; i++)
+        {
+          icon_info = theme_lookup_icon (theme, icon_names[i], size, allow_svg, use_builtin);
+          if (icon_info)
+            goto out;
+        }
+    }
+
+  for (i = 0; icon_names[i]; i++)
+    {
+      unthemed_icon = g_hash_table_lookup (priv->unthemed_icons, icon_names[i]);
+      if (unthemed_icon)
+        break;
     }
 
-  unthemed_icon = g_hash_table_lookup (priv->unthemed_icons, icon_name);
   if (unthemed_icon)
     {
       icon_info = icon_info_new ();
@@ -1315,7 +1322,7 @@ gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
 			   "was not found either, perhaps you need to install it.\n"
 			   "You can get a copy from:\n"
 			   "\t%s"),
-			 icon_name, DEFAULT_THEME_NAME, "http://icon-theme.freedesktop.org/releases");
+			 icon_names[0], DEFAULT_THEME_NAME, "http://icon-theme.freedesktop.org/releases");
 	    }
 	}
     }
@@ -1323,6 +1330,117 @@ gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
   return icon_info;
 }
 
+
+/**
+ * gtk_icon_theme_lookup_icon:
+ * @icon_theme: a #GtkIconTheme
+ * @icon_name: the name of the icon to lookup
+ * @size: desired icon size
+ * @flags: flags modifying the behavior of the icon lookup
+ * 
+ * Looks up a named icon and returns a structure containing
+ * information such as the filename of the icon. The icon
+ * can then be rendered into a pixbuf using
+ * gtk_icon_info_load_icon(). (gtk_icon_theme_load_icon()
+ * combines these two steps if all you need is the pixbuf.)
+ * 
+ * Return value: a #GtkIconInfo structure containing information
+ * about the icon, or %NULL if the icon wasn't found. Free with
+ * gtk_icon_info_free()
+ *
+ * Since: 2.4
+ */
+GtkIconInfo *
+gtk_icon_theme_lookup_icon (GtkIconTheme       *icon_theme,
+			    const gchar        *icon_name,
+			    gint                size,
+			    GtkIconLookupFlags  flags)
+{
+  GtkIconInfo *info;
+
+  g_return_val_if_fail (GTK_IS_ICON_THEME (icon_theme), NULL);
+  g_return_val_if_fail (icon_name != NULL, NULL);
+  g_return_val_if_fail ((flags & GTK_ICON_LOOKUP_NO_SVG) == 0 ||
+			(flags & GTK_ICON_LOOKUP_FORCE_SVG) == 0, NULL);
+
+  GTK_NOTE (ICONTHEME, 
+	    g_print ("gtk_icon_theme_lookup_icon %s\n", icon_name));
+
+  if (flags & GTK_ICON_LOOKUP_GENERIC_FALLBACK)
+    {
+      gchar **names;
+      gint dashes, i;
+      gchar *p;
+ 
+      dashes = 0;
+      for (p = icon_name; *p; p++)
+        if (*p == '-')
+          dashes++;
+
+      names = g_new (gchar *, dashes + 2);
+      names[0] = g_strdup (icon_name);
+      for (i = 1; i <= dashes; i++)
+        {
+          names[i] = g_strdup (names[i - 1]);
+          p = strrchr (names[i], '-');
+          *p = '\0';
+        }
+      names[dashes + 1] = NULL;
+   
+      info = choose_icon (icon_theme, names, size, flags);
+      
+      g_strfreev (names);
+    }
+  else 
+    {
+      gchar *names[2];
+      
+      names[0] = icon_name;
+      names[1] = NULL;
+
+      info = choose_icon (icon_theme, names, size, flags);
+    }
+
+  return info;
+}
+
+/**
+ * gtk_icon_theme_choose_icon:
+ * @icon_theme: a #GtkIconTheme
+ * @icon_names: %NULL-terminated array of icon names to lookup
+ * @size: desired icon size
+ * @flags: flags modifying the behavior of the icon lookup
+ * 
+ * Looks up a named icon and returns a structure containing
+ * information such as the filename of the icon. The icon
+ * can then be rendered into a pixbuf using
+ * gtk_icon_info_load_icon(). (gtk_icon_theme_load_icon()
+ * combines these two steps if all you need is the pixbuf.)
+ *
+ * If @icon_names contains more than one name, this function 
+ * tries them all in the given order before falling back to 
+ * inherited icon themes.
+ * 
+ * Return value: a #GtkIconInfo structure containing information
+ * about the icon, or %NULL if the icon wasn't found. Free with
+ * gtk_icon_info_free()
+ *
+ * Since: 2.12
+ */
+GtkIconInfo *
+gtk_icon_theme_choose_icon (GtkIconTheme       *icon_theme,
+			    const gchar        *icon_names[],
+			    gint                size,
+			    GtkIconLookupFlags  flags)
+{
+  g_return_val_if_fail (GTK_IS_ICON_THEME (icon_theme), NULL);
+  g_return_val_if_fail (icon_names != NULL, NULL);
+  g_return_val_if_fail ((flags & GTK_ICON_LOOKUP_NO_SVG) == 0 ||
+			(flags & GTK_ICON_LOOKUP_FORCE_SVG) == 0, NULL);
+
+  return choose_icon (icon_theme, icon_names, size, flags);
+}
+
 /* Error quark */
 GQuark
 gtk_icon_theme_error_quark (void)
@@ -1627,6 +1745,51 @@ gtk_icon_theme_list_icons (GtkIconTheme *icon_theme,
   return list;
 }
 
+/**
+ * gtk_icon_theme_list_contexts:
+ * @icon_theme: a #GtkIconTheme
+ *
+ * Gets the list of contexts available within the current
+ * hierarchy of icon themes
+ *
+ * Return value: a #GList list holding the names of all the
+ *  contexts in the theme. You must first free each element
+ *  in the list with g_free(), then free the list itself
+ *  with g_list_free().
+ *
+ * Since: 2.12
+ **/
+GList *
+gtk_icon_theme_list_contexts (GtkIconTheme *icon_theme)
+{
+  GtkIconThemePrivate *priv;
+  GHashTable *contexts;
+  GList *list, *l;
+
+  priv = icon_theme->priv;
+  
+  ensure_valid_themes (icon_theme);
+
+  contexts = g_hash_table_new (g_str_hash, g_str_equal);
+
+  l = priv->themes;
+  while (l != NULL)
+    {
+      theme_list_contexts (l->data, contexts);
+      l = l->next;
+    }
+
+  list = NULL;
+
+  g_hash_table_foreach (contexts,
+			add_key_to_list,
+			&list);
+
+  g_hash_table_destroy (contexts);
+
+  return list;
+}
+
 /**
  * gtk_icon_theme_get_example_icon_name:
  * @icon_theme: a #GtkIconTheme
@@ -1666,6 +1829,44 @@ gtk_icon_theme_get_example_icon_name (GtkIconTheme *icon_theme)
   return NULL;
 }
 
+
+static gboolean
+rescan_themes (GtkIconTheme *icon_theme)
+{
+  GtkIconThemePrivate *priv;
+  IconThemeDirMtime *dir_mtime;
+  GList *d;
+  int stat_res;
+  struct stat stat_buf;
+  GTimeVal tv;
+
+  priv = icon_theme->priv;
+
+  for (d = priv->dir_mtimes; d != NULL; d = d->next)
+    {
+      dir_mtime = d->data;
+
+      stat_res = g_stat (dir_mtime->dir, &stat_buf);
+
+      /* dir mtime didn't change */
+      if (stat_res == 0 &&
+	  S_ISDIR (stat_buf.st_mode) &&
+	  dir_mtime->mtime == stat_buf.st_mtime)
+	continue;
+      /* didn't exist before, and still doesn't */
+      if (dir_mtime->mtime == 0 &&
+	  (stat_res != 0 || !S_ISDIR (stat_buf.st_mode)))
+	continue;
+
+      return TRUE;
+    }
+
+  g_get_current_time (&tv);
+  priv->last_stat_time = tv.tv_sec;
+
+  return FALSE;
+}
+
 /**
  * gtk_icon_theme_rescan_if_needed:
  * @icon_theme: a #GtkIconTheme
@@ -1682,41 +1883,15 @@ gtk_icon_theme_get_example_icon_name (GtkIconTheme *icon_theme)
 gboolean
 gtk_icon_theme_rescan_if_needed (GtkIconTheme *icon_theme)
 {
-  GtkIconThemePrivate *priv;
-  IconThemeDirMtime *dir_mtime;
-  GList *d;
-  int stat_res;
-  struct stat stat_buf;
-  GTimeVal tv;
+  gboolean retval;
 
   g_return_val_if_fail (GTK_IS_ICON_THEME (icon_theme), FALSE);
 
-  priv = icon_theme->priv;
-  
-  for (d = priv->dir_mtimes; d != NULL; d = d->next)
-    {
-      dir_mtime = d->data;
-
-      stat_res = g_stat (dir_mtime->dir, &stat_buf);
-
-      /* dir mtime didn't change */
-      if (stat_res == 0 && 
-	  S_ISDIR (stat_buf.st_mode) &&
-	  dir_mtime->mtime == stat_buf.st_mtime)
-	continue;
-      /* didn't exist before, and still doesn't */
-      if (dir_mtime->mtime == 0 &&
-	  (stat_res != 0 || !S_ISDIR (stat_buf.st_mode)))
-	continue;
-	  
+  retval = rescan_themes (icon_theme);
+  if (retval)
       do_theme_change (icon_theme);
-      return TRUE;
-    }
-  
-  g_get_current_time (&tv);
-  priv->last_stat_time = tv.tv_sec;
 
-  return FALSE;
+  return retval;
 }
 
 static void
@@ -1855,7 +2030,7 @@ theme_dir_get_icon_suffix (IconThemeDir *dir,
     suffix = GPOINTER_TO_UINT (g_hash_table_lookup (dir->icons, icon_name));
 
   GTK_NOTE (ICONTHEME, 
-	    g_print ("get_icon_suffix%s %d\n", dir->cache ? " (cached)" : "", suffix));
+	    g_print ("get_icon_suffix%s %u\n", dir->cache ? " (cached)" : "", suffix));
 
   return suffix;
 }
@@ -1872,12 +2047,13 @@ theme_lookup_icon (IconTheme          *theme,
   char *file;
   int min_difference, difference;
   BuiltinIcon *closest_builtin = NULL;
-  gboolean smaller, has_larger;
+  gboolean smaller, has_larger, match;
   IconSuffix suffix;
 
   min_difference = G_MAXINT;
   min_dir = NULL;
   has_larger = FALSE;
+  match = FALSE;
 
   /* Builtin icons are logically part of the default theme and
    * are searched before other subdirectories of the default theme.
@@ -1911,31 +2087,55 @@ theme_lookup_icon (IconTheme          *theme,
 
 	  if (difference == 0)
 	    {
-	      min_dir = dir;
-	      break;
+              if (dir->type == ICON_THEME_DIR_SCALABLE)
+                {
+                  /* don't pick scalable if we already found
+                   * a matching non-scalable dir
+                   */
+                  if (!match)
+                    {
+	              min_dir = dir;
+	              break;
+                    }
+                }
+              else
+                {
+                  /* for a matching non-scalable dir keep
+                   * going and look for a closer match
+                   */             
+                  difference = abs (size - dir->size);
+                  if (!match || difference < min_difference)
+                    {
+                      match = TRUE;
+                      min_difference = difference;
+	              min_dir = dir;
+                    }
+                  if (difference == 0)
+                    break;
+                }
+	    } 
+  
+          if (!match)
+            {
+	      if (!has_larger)
+	        {
+	          if (difference < min_difference || smaller)
+	  	    {
+		      min_difference = difference;
+		      min_dir = dir;
+		      has_larger = smaller;
+	 	    }
+	        }
+	      else
+	        {
+	          if (difference < min_difference && smaller)
+		    {
+		      min_difference = difference;
+		      min_dir = dir;
+		    }
+	        }
 	    }
-
-	  if (!has_larger)
-	    {
-	      if (difference < min_difference || smaller)
-		{
-		  min_difference = difference;
-		  min_dir = dir;
-		  closest_builtin = NULL;
-		  has_larger = smaller;
-		}
-	    }
-	  else
-	    {
-	      if (difference < min_difference && smaller)
-		{
-		  min_difference = difference;
-		  min_dir = dir;
-		  closest_builtin = NULL;
-		}
-	    }
-
-	}
+        }
 
       l = l->next;
 
@@ -1946,9 +2146,6 @@ theme_lookup_icon (IconTheme          *theme,
 	}
     }
 
-  if (closest_builtin)
-    return icon_info_new_builtin (closest_builtin);
-  
   if (min_dir)
     {
       GtkIconInfo *icon_info = icon_info_new ();
@@ -1958,13 +2155,23 @@ theme_lookup_icon (IconTheme          *theme,
       suffix = best_suffix (suffix, allow_svg);
       g_assert (suffix != ICON_SUFFIX_NONE);
       
-      file = g_strconcat (icon_name, string_from_suffix (suffix), NULL);
-      icon_info->filename = g_build_filename (min_dir->dir, file, NULL);
-      g_free (file);
+      if (min_dir->dir)
+        {
+          file = g_strconcat (icon_name, string_from_suffix (suffix), NULL);
+          icon_info->filename = g_build_filename (min_dir->dir, file, NULL);
+          g_free (file);
 #ifdef G_OS_WIN32
-      icon_info->cp_filename = g_locale_from_utf8 (icon_info->filename,
+          icon_info->cp_filename = g_locale_from_utf8 (icon_info->filename,
 						   -1, NULL, NULL, NULL);
 #endif
+        }
+      else
+        {
+          icon_info->filename = NULL;
+#ifdef G_OS_WIN32
+          icon_info->cp_filename = NULL;
+#endif
+        }
       
       if (min_dir->icon_data != NULL)
 	icon_info->data = g_hash_table_lookup (min_dir->icon_data, icon_name);
@@ -2014,7 +2221,10 @@ theme_lookup_icon (IconTheme          *theme,
       
       return icon_info;
     }
- 
+
+  if (closest_builtin)
+    return icon_info_new_builtin (closest_builtin);
+  
   return NULL;
 }
 
@@ -2051,6 +2261,25 @@ theme_list_icons (IconTheme  *theme,
     }
 }
 
+static void
+theme_list_contexts (IconTheme  *theme, 
+		     GHashTable *contexts)
+{
+  GList *l = theme->dirs;
+  IconThemeDir *dir;
+  const char *context;
+
+  while (l != NULL)
+    {
+      dir = l->data;
+
+      context = g_quark_to_string (dir->context);
+      g_hash_table_replace (contexts, context, NULL);
+
+      l = l->next;
+    }
+}
+
 static void
 load_icon_data (IconThemeDir *dir, const char *path, const char *name)
 {
@@ -2177,8 +2406,8 @@ scan_directory (GtkIconThemePrivate *icon_theme,
       base_name = strip_suffix (name);
 
       hash_suffix = GPOINTER_TO_INT (g_hash_table_lookup (dir->icons, base_name));
+      g_hash_table_replace (icon_theme->all_icons, base_name, NULL);
       g_hash_table_replace (dir->icons, base_name, GUINT_TO_POINTER (hash_suffix| suffix));
-      g_hash_table_insert (icon_theme->all_icons, base_name, NULL);
     }
   
   g_dir_close (gdir);
@@ -2400,11 +2629,9 @@ gtk_icon_info_free (GtkIconInfo *icon_info)
 {
   g_return_if_fail (icon_info != NULL);
 
-  if (icon_info->filename)
-    g_free (icon_info->filename);
+  g_free (icon_info->filename);
 #ifdef G_OS_WIN32
-  if (icon_info->cp_filename)
-    g_free (icon_info->cp_filename);
+  g_free (icon_info->cp_filename);
 #endif
   if (icon_info->pixbuf)
     g_object_unref (icon_info->pixbuf);
@@ -2526,8 +2753,7 @@ load_svg_at_size (const gchar *filename,
  bail:
   if (loader)
     g_object_unref (loader);
-  if (contents)
-    g_free (contents);
+  g_free (contents);
   
   return pixbuf;
 }
@@ -2597,7 +2823,7 @@ icon_info_ensure_scale_and_pixbuf (GtkIconInfo *icon_info,
     return TRUE;
 
   /* At this point, we need to actually get the icon; either from the
-   * builting image or by loading the file
+   * builtin image or by loading the file
    */
   if (icon_info->cache_pixbuf)
     source_pixbuf = g_object_ref (icon_info->cache_pixbuf);
@@ -2632,7 +2858,7 @@ icon_info_ensure_scale_and_pixbuf (GtkIconInfo *icon_info,
    * the job. This is a bit of a waste when we scale here
    * and never get the final pixbuf; at the cost of a bit of
    * extra complexity, we could keep the source pixbuf around
-   * but not actually scale it until neede.
+   * but not actually scale it until needed.
    */
     
   if (icon_info->scale == 1.0)
