@@ -30,6 +30,7 @@
 #include <stdlib.h>       
 #include <fcntl.h>
 
+#include <glib/gstdio.h>
 #include "gtkprintoperation-private.h"
 #include "gtkmarshal.h"
 #include "gtkmessagedialog.h"
@@ -44,8 +45,8 @@
 #include "gtkprinter-private.h"
 #include "gtkprintjob.h"
 #include "gtklabel.h"
-#include "gtkalias.h"
 #include "gtkintl.h"
+#include "gtkalias.h"
 
 typedef struct 
 {
@@ -118,20 +119,24 @@ op_unix_free (GtkPrintOperationUnix *op_unix)
 
 static gchar *
 shell_command_substitute_file (const gchar *cmd,
-			       const gchar *filename)
+			       const gchar *pdf_filename,
+			       const gchar *settings_filename,
+                               gboolean    *pdf_filename_replaced,
+                               gboolean    *settings_filename_replaced)
 {
   const gchar *inptr, *start;
-  gchar *result;
   GString *final;
 
   g_return_val_if_fail (cmd != NULL, NULL);
-  g_return_val_if_fail (filename != NULL, NULL);
+  g_return_val_if_fail (pdf_filename != NULL, NULL);
+  g_return_val_if_fail (settings_filename != NULL, NULL);
 
-  result = NULL;
   final = g_string_new (NULL);
 
-  start = inptr = cmd;
+  *pdf_filename_replaced = FALSE;
+  *settings_filename_replaced = FALSE;
 
+  start = inptr = cmd;
   while ((inptr = strchr (inptr, '%')) != NULL) 
     {
       g_string_append_len (final, start, inptr - start);
@@ -139,7 +144,13 @@ shell_command_substitute_file (const gchar *cmd,
       switch (*inptr) 
         {
           case 'f':
-            g_string_append (final, filename ? filename : "");
+            g_string_append (final, pdf_filename);
+            *pdf_filename_replaced = TRUE;
+            break;
+
+          case 's':
+            g_string_append (final, settings_filename);
+            *settings_filename_replaced = TRUE;
             break;
 
           case '%':
@@ -158,39 +169,7 @@ shell_command_substitute_file (const gchar *cmd,
     }
   g_string_append (final, start);
 
-  result = final->str;
-
-  g_string_free (final, FALSE);
-
-  return result;
-}
-
-static void
-gtk_print_operation_unix_initialize (void)
-{
-  static gboolean initialized = FALSE;
-
-  if (!initialized)
-    {
-      /**
-       * GtkSettings::gtk-print-preview-command:
-       *
-       * A command to run for displaying the print preview. The command
-       * should contain a %f placeholder, which will get replaced by
-       * the path to the pdf file.
-       *
-       * The preview application is responsible for removing the pdf file
-       * when it is done.
-       *
-       * Since: 2.10
-       */
-      gtk_settings_install_property (g_param_spec_string ("gtk-print-preview-command",
-							  P_("Default command to run when displaying a print preview"),
-							  P_("Command to run when displaying a print preview"),
-							  GTK_PRINT_PREVIEW_COMMAND,
-							  GTK_PARAM_READWRITE)); 
-      initialized = TRUE;
-    }
+  return g_string_free (final, FALSE);
 }
 
 void
@@ -204,31 +183,55 @@ _gtk_print_operation_platform_backend_launch_preview (GtkPrintOperation *op,
   gchar *cmd;
   gchar *preview_cmd;
   GtkSettings *settings;
+  GtkPrintSettings *print_settings;
+  gchar *settings_filename = NULL;
   gchar *quoted_filename;
+  gchar *quoted_settings_filename;
+  gboolean filename_used = FALSE;
+  gboolean settings_used = FALSE;
   GdkScreen *screen;
   GError *error = NULL;
-
-  gtk_print_operation_unix_initialize ();
+  gint fd;
+  gboolean retval;
 
   cairo_surface_destroy (surface);
  
-  settings = gtk_settings_get_default ();
-  g_object_get (settings, "gtk-print-preview-command", &preview_cmd, NULL);
-
-  quoted_filename = g_shell_quote (filename);
-  cmd = shell_command_substitute_file (preview_cmd, quoted_filename);
-  g_shell_parse_argv (cmd, &argc, &argv, &error);
-
-  if (error != NULL)
-    goto out;
-
   if (parent)
     screen = gtk_window_get_screen (parent);
   else
     screen = gdk_screen_get_default ();
-  
+
+  fd = g_file_open_tmp ("settingsXXXXXX.ini", &settings_filename, &error);
+  if (fd < 0) 
+    goto out;
+
+  print_settings = gtk_print_operation_get_print_settings (op);
+  retval = gtk_print_settings_to_file (print_settings, settings_filename, &error);
+  close (fd);
+
+  if (!retval)
+    goto out;
+
+  settings = gtk_settings_get_for_screen (screen);
+  g_object_get (settings, "gtk-print-preview-command", &preview_cmd, NULL);
+
+  quoted_filename = g_shell_quote (filename);
+  quoted_settings_filename = g_shell_quote (settings_filename);
+  cmd = shell_command_substitute_file (preview_cmd, quoted_filename, quoted_settings_filename, &filename_used, &settings_used);
+  g_shell_parse_argv (cmd, &argc, &argv, &error);
+
+  g_free (preview_cmd);
+  g_free (quoted_filename);
+  g_free (quoted_settings_filename);
+  g_free (cmd);
+
+  if (error != NULL)
+    goto out;
+
   gdk_spawn_on_screen (screen, NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, &error);
 
+  g_strfreev (argv);
+
  out:
   if (error != NULL)
     {
@@ -240,19 +243,24 @@ _gtk_print_operation_platform_backend_launch_preview (GtkPrintOperation *op,
                                         _("Error launching preview") /* FIXME better text */);
       gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (edialog),
                                                 "%s", error->message);
-      gtk_window_set_modal (GTK_WINDOW (edialog), TRUE);
       g_signal_connect (edialog, "response",
                         G_CALLBACK (gtk_widget_destroy), NULL);
 
       gtk_window_present (GTK_WINDOW (edialog));
 
-      g_error_free (error); 
+      g_error_free (error);
+
+      filename_used = FALSE; 
+      settings_used = FALSE;
    } 
 
-  g_free (cmd);
-  g_free (quoted_filename);
-  g_free (preview_cmd);
-  g_strfreev (argv);
+  if (!filename_used)
+    g_unlink (filename);
+
+  if (!settings_used)
+    g_unlink (settings_filename);
+
+  g_free (settings_filename);
 }
 
 static void
@@ -310,6 +318,7 @@ unix_end_run (GtkPrintOperation *op,
 
   if (wait)
     {
+      g_object_ref (op);
       if (!op_unix->data_sent)
 	{
 	  GDK_THREADS_LEAVE ();  
@@ -317,6 +326,8 @@ unix_end_run (GtkPrintOperation *op,
 	  GDK_THREADS_ENTER ();  
 	}
       g_main_loop_unref (op_unix->loop);
+      op_unix->loop = NULL;
+      g_object_unref (op);
     }
 }
 
@@ -345,9 +356,7 @@ get_print_dialog (GtkPrintOperation *op,
 						 GTK_PRINT_CAPABILITY_COLLATE |
 						 GTK_PRINT_CAPABILITY_REVERSE |
 						 GTK_PRINT_CAPABILITY_SCALE |
-                                                 GTK_PRINT_CAPABILITY_GENERATE_PDF |
-						 GTK_PRINT_CAPABILITY_GENERATE_PS |
-                                                 GTK_PRINT_CAPABILITY_PREVIEW);
+						 GTK_PRINT_CAPABILITY_PREVIEW);
 
   if (priv->print_settings)
     gtk_print_unix_dialog_set_settings (GTK_PRINT_UNIX_DIALOG (pd),
@@ -579,6 +588,7 @@ _gtk_print_operation_platform_backend_run_dialog_async (GtkPrintOperation
   rdata = g_new (PrintResponseData, 1);
   rdata->op = g_object_ref (op);
   rdata->do_print = FALSE;
+  rdata->do_preview = FALSE;
   rdata->result = GTK_PRINT_OPERATION_RESULT_CANCEL;
   rdata->print_cb = print_cb;
   rdata->parent = parent;
@@ -658,8 +668,6 @@ _gtk_print_operation_platform_backend_create_preview_surface (GtkPrintOperation
   fd = g_mkstemp (filename);
   *target = filename;
   
-  g_print ("target is %s\n", filename);
-
   paper_size = gtk_page_setup_get_paper_size (page_setup);
   w = gtk_paper_size_get_width (paper_size, GTK_UNIT_POINTS);
   h = gtk_paper_size_get_height (paper_size, GTK_UNIT_POINTS);
@@ -715,6 +723,7 @@ _gtk_print_operation_platform_backend_run_dialog (GtkPrintOperation *op,
    
   rdata.op = op;
   rdata.do_print = FALSE;
+  rdata.do_preview = FALSE;
   rdata.result = GTK_PRINT_OPERATION_RESULT_CANCEL;
   rdata.print_cb = NULL;
   rdata.destroy = NULL;
