@@ -3,6 +3,7 @@
  * Copyright (C) 2003 Sun Microsystems, Inc.
  * Copyright (C) 2005 Hans Breuer <hans@breuer.org>
  * Copyright (C) 2005 Novell, Inc.
+ * Copyright (C) 2006 Imendio AB
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -23,6 +24,7 @@
  *	Mark McLoughlin <mark@skynet.ie>
  *	Hans Breuer <hans@breuer.org>
  *	Tor Lillqvist <tml@novell.com>
+ *	Mikael Hallendal <micke@imendio.com>
  */
 
 #include <config.h>
@@ -39,8 +41,6 @@
 #include "gtkprivate.h"
 #include "gtkwidget.h"
 
-#include "gtkalias.h"
-
 #ifdef GDK_WINDOWING_WIN32
 #include "gtkicontheme.h"
 #include "gtklabel.h"
@@ -49,6 +49,13 @@
 #define WM_GTK_TRAY_NOTIFICATION (WM_USER+1)
 #endif
 
+#ifdef GDK_WINDOWING_QUARTZ
+#include "gtkicontheme.h"
+#include "gtklabel.h"
+#endif	
+
+#include "gtkalias.h"
+
 #define BLINK_TIMEOUT 500
 
 enum
@@ -60,7 +67,10 @@ enum
   PROP_ICON_NAME,
   PROP_STORAGE_TYPE,
   PROP_SIZE,
+  PROP_SCREEN,
   PROP_VISIBLE,
+  PROP_ORIENTATION,
+  PROP_EMBEDDED,
   PROP_BLINKING
 };
 
@@ -72,6 +82,12 @@ enum
   LAST_SIGNAL
 };
 
+static guint status_icon_signals [LAST_SIGNAL] = { 0 };
+
+#ifdef GDK_WINDOWING_QUARTZ
+#include "gtkstatusicon-quartz.c"
+#endif
+
 struct _GtkStatusIconPrivate
 {
 #ifdef GDK_WINDOWING_X11
@@ -83,6 +99,13 @@ struct _GtkStatusIconPrivate
 #ifdef GDK_WINDOWING_WIN32
   GtkWidget     *dummy_widget;
   NOTIFYICONDATAW nid;
+  gint		last_click_x, last_click_y;
+  GtkOrientation orientation;
+#endif
+	
+#ifdef GDK_WINDOWING_QUARTZ
+  GtkWidget     *dummy_widget;
+  GtkQuartzStatusIcon *status_item;
 #endif
 
   gint          size;
@@ -107,6 +130,9 @@ struct _GtkStatusIconPrivate
   guint         visible : 1;
 };
 
+static GObject* gtk_status_icon_constructor      (GType                  type,
+                                                  guint                  n_construct_properties,
+                                                  GObjectConstructParam *construct_params);
 static void     gtk_status_icon_finalize         (GObject        *object);
 static void     gtk_status_icon_set_property     (GObject        *object,
 						  guint           prop_id,
@@ -120,6 +146,11 @@ static void     gtk_status_icon_get_property     (GObject        *object,
 #ifdef GDK_WINDOWING_X11
 static void     gtk_status_icon_size_allocate    (GtkStatusIcon  *status_icon,
 						  GtkAllocation  *allocation);
+static void     gtk_status_icon_screen_changed   (GtkStatusIcon  *status_icon,
+						  GdkScreen      *old_screen);
+static void     gtk_status_icon_embedded_changed (GtkStatusIcon *status_icon);
+static void     gtk_status_icon_orientation_changed (GtkStatusIcon *status_icon);
+
 #endif
 static gboolean gtk_status_icon_button_press     (GtkStatusIcon  *status_icon,
 						  GdkEventButton *event);
@@ -127,8 +158,6 @@ static void     gtk_status_icon_disable_blinking (GtkStatusIcon  *status_icon);
 static void     gtk_status_icon_reset_image_data (GtkStatusIcon  *status_icon);
 					   
 
-static guint status_icon_signals [LAST_SIGNAL] = { 0 };
-
 G_DEFINE_TYPE (GtkStatusIcon, gtk_status_icon, G_TYPE_OBJECT)
 
 static void
@@ -136,6 +165,7 @@ gtk_status_icon_class_init (GtkStatusIconClass *class)
 {
   GObjectClass *gobject_class = (GObjectClass *) class;
 
+  gobject_class->constructor  = gtk_status_icon_constructor;
   gobject_class->finalize     = gtk_status_icon_finalize;
   gobject_class->set_property = gtk_status_icon_set_property;
   gobject_class->get_property = gtk_status_icon_get_property;
@@ -191,6 +221,14 @@ gtk_status_icon_class_init (GtkStatusIconClass *class)
 						     0,
 						     GTK_PARAM_READABLE));
 
+  g_object_class_install_property (gobject_class,
+				   PROP_SCREEN,
+				   g_param_spec_object ("screen",
+ 							P_("Screen"),
+ 							P_("The screen where this status icon will be displayed"),
+							GDK_TYPE_SCREEN,
+ 							GTK_PARAM_READWRITE));
+
   g_object_class_install_property (gobject_class,
 				   PROP_BLINKING,
 				   g_param_spec_boolean ("blinking",
@@ -208,6 +246,39 @@ gtk_status_icon_class_init (GtkStatusIconClass *class)
 							 GTK_PARAM_READWRITE));
 
 
+  /**
+   * GtkStatusIcon:embedded: 
+   *
+   * %TRUE if the statusicon is embedded in a notification area.
+   *
+   * Since: 2.12
+   */
+  g_object_class_install_property (gobject_class,
+				   PROP_EMBEDDED,
+				   g_param_spec_boolean ("embedded",
+							 P_("Embedded"),
+							 P_("Whether or not the status icon is embedded"),
+							 FALSE,
+							 GTK_PARAM_READABLE));
+
+  /**
+   * GtkStatusIcon:orientation:
+   *
+   * The orientation of the tray in which the statusicon 
+   * is embedded. 
+   *
+   * Since: 2.12
+   */
+  g_object_class_install_property (gobject_class,
+				   PROP_ORIENTATION,
+				   g_param_spec_enum ("orientation",
+						      P_("Orientation"),
+						      P_("The orientation of the tray"),
+						      GTK_TYPE_ORIENTATION,
+						      GTK_ORIENTATION_HORIZONTAL,
+						      GTK_PARAM_READABLE));
+
+
   /**
    * GtkStatusIcon::activate:
    * @status_icon: the object which received the signal
@@ -289,9 +360,10 @@ gtk_status_icon_class_init (GtkStatusIconClass *class)
 #ifdef GDK_WINDOWING_WIN32
 
 static void
-build_button_event (GdkEventButton *e,
-		    GdkEventType    type,
-		    guint           button)
+build_button_event (GtkStatusIconPrivate *priv,
+		    GdkEventButton       *e,
+		    GdkEventType          type,
+		    guint                 button)
 {
   POINT pos;
   GdkRectangle monitor0;
@@ -303,8 +375,8 @@ build_button_event (GdkEventButton *e,
   e->send_event = TRUE;
   e->time = GetTickCount ();
   GetCursorPos (&pos);
-  e->x = pos.x + monitor0.x;
-  e->y = pos.y + monitor0.y;
+  priv->last_click_x = e->x = pos.x + monitor0.x;
+  priv->last_click_y = e->y = pos.y + monitor0.y;
   e->axes = NULL;
   e->state = 0;
   e->button = button;
@@ -328,7 +400,7 @@ wndproc (HWND   hwnd,
 	{
 	case WM_LBUTTONDOWN:
 	case WM_RBUTTONDOWN:
-	  build_button_event (&e, GDK_BUTTON_PRESS,
+	  build_button_event (status_icon->priv, &e, GDK_BUTTON_PRESS,
 			      (lparam == WM_LBUTTONDOWN) ? 1 : 3);
 	  gtk_status_icon_button_press (status_icon, &e);
 	  break;
@@ -400,38 +472,40 @@ gtk_status_icon_init (GtkStatusIcon *status_icon)
   gtk_widget_add_events (GTK_WIDGET (priv->tray_icon),
 			 GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
 
+  g_signal_connect_swapped (priv->tray_icon, "notify::embedded",
+			    G_CALLBACK (gtk_status_icon_embedded_changed), status_icon);
+  g_signal_connect_swapped (priv->tray_icon, "notify::orientation",
+			    G_CALLBACK (gtk_status_icon_orientation_changed), status_icon);
   g_signal_connect_swapped (priv->tray_icon, "button-press-event",
 			    G_CALLBACK (gtk_status_icon_button_press), status_icon);
+  g_signal_connect_swapped (priv->tray_icon, "screen-changed",
+		    	    G_CALLBACK (gtk_status_icon_screen_changed), status_icon);
   priv->image = gtk_image_new ();
   gtk_container_add (GTK_CONTAINER (priv->tray_icon), priv->image);
+  gtk_widget_show (priv->image);
 
   g_signal_connect_swapped (priv->image, "size-allocate",
 			    G_CALLBACK (gtk_status_icon_size_allocate), status_icon);
 
-  gtk_widget_show (priv->image);
-  gtk_widget_show (priv->tray_icon);
-
   status_icon->priv->tooltips = gtk_tooltips_new ();
   g_object_ref_sink (priv->tooltips);
 #endif
 
 #ifdef GDK_WINDOWING_WIN32
 
-  /* Code to get position and orientation of Windows taskbar. Not needed
-   * currently, kept for reference.
-   */
-#if 0
+  /* Get position and orientation of Windows taskbar. */
   {
     APPBARDATA abd;
     
     abd.cbSize = sizeof (abd);
     SHAppBarMessage (ABM_GETTASKBARPOS, &abd);
     if (abd.rc.bottom - abd.rc.top > abd.rc.right - abd.rc.left)
-      orientation = GTK_ORIENTATION_VERTICAL;
+      priv->orientation = GTK_ORIENTATION_VERTICAL;
     else
-      orientation = GTK_ORIENTATION_HORIZONTAL;
+      priv->orientation = GTK_ORIENTATION_HORIZONTAL;
   }
-#endif
+
+  priv->last_click_x = priv->last_click_y = 0;
 
   /* Are the system tray icons always 16 pixels square? */
   priv->size         = 16;
@@ -453,6 +527,44 @@ gtk_status_icon_init (GtkStatusIcon *status_icon)
       priv->nid.hWnd = NULL;
     }
 #endif
+	
+#ifdef GDK_WINDOWING_QUARTZ
+  priv->dummy_widget = gtk_label_new ("");
+
+  QUARTZ_POOL_ALLOC;
+
+  priv->status_item = [[GtkQuartzStatusIcon alloc] initWithStatusIcon:status_icon];
+
+  priv->image_width = priv->image_height = [priv->status_item getHeight];
+  priv->size = priv->image_height;
+
+  QUARTZ_POOL_RELEASE;
+
+#endif 
+}
+
+static GObject*
+gtk_status_icon_constructor (GType                  type,
+                             guint                  n_construct_properties,
+                             GObjectConstructParam *construct_params)
+{
+  GObject *object;
+  GtkStatusIcon *status_icon;
+  GtkStatusIconPrivate *priv;
+  
+  object = G_OBJECT_CLASS (gtk_status_icon_parent_class)->constructor (type,
+                                                                       n_construct_properties,
+                                                                       construct_params);
+
+#ifdef GDK_WINDOWING_X11
+  status_icon = GTK_STATUS_ICON (object);
+  priv = status_icon->priv;
+  
+  if (priv->visible)
+    gtk_widget_show (priv->tray_icon);
+#endif
+
+  return object;
 }
 
 static void
@@ -480,9 +592,17 @@ gtk_status_icon_finalize (GObject *object)
 #ifdef GDK_WINDOWING_WIN32
   if (priv->nid.hWnd != NULL && priv->visible)
     Shell_NotifyIconW (NIM_DELETE, &priv->nid);
+  if (priv->nid.hIcon)
+    DestroyIcon (priv->nid.hIcon);
 
   gtk_widget_destroy (priv->dummy_widget);
 #endif
+	
+#ifdef GDK_WINDOWING_QUARTZ
+  QUARTZ_POOL_ALLOC;
+  [priv->status_item release];
+  QUARTZ_POOL_RELEASE;
+#endif
 
   G_OBJECT_CLASS (gtk_status_icon_parent_class)->finalize (object);
 }
@@ -509,6 +629,9 @@ gtk_status_icon_set_property (GObject      *object,
     case PROP_ICON_NAME:
       gtk_status_icon_set_from_icon_name (status_icon, g_value_get_string (value));
       break;
+    case PROP_SCREEN:
+      gtk_status_icon_set_screen (status_icon, g_value_get_object (value));
+      break;
     case PROP_BLINKING:
       gtk_status_icon_set_blinking (status_icon, g_value_get_boolean (value));
       break;
@@ -562,12 +685,26 @@ gtk_status_icon_get_property (GObject    *object,
     case PROP_SIZE:
       g_value_set_int (value, gtk_status_icon_get_size (status_icon));
       break;
+    case PROP_SCREEN:
+      g_value_set_object (value, gtk_status_icon_get_screen (status_icon));
+      break;
     case PROP_BLINKING:
       g_value_set_boolean (value, gtk_status_icon_get_blinking (status_icon));
       break;
     case PROP_VISIBLE:
       g_value_set_boolean (value, gtk_status_icon_get_visible (status_icon));
       break;
+    case PROP_EMBEDDED:
+      g_value_set_boolean (value, gtk_status_icon_is_embedded (status_icon));
+      break;
+    case PROP_ORIENTATION:
+#ifdef GDK_WINDOWING_X11
+      g_value_set_enum (value, _gtk_tray_icon_get_orientation (GTK_TRAY_ICON (status_icon->priv->tray_icon)));
+#endif
+#ifdef GDK_WINDOWING_WIN32
+      g_value_set_enum (value, status_icon->priv->orientation);
+#endif
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -783,6 +920,9 @@ static void
 gtk_status_icon_update_image (GtkStatusIcon *status_icon)
 {
   GtkStatusIconPrivate *priv = status_icon->priv;
+#ifdef GDK_WINDOWING_WIN32
+  HICON prev_hicon;
+#endif
 
   if (priv->blink_off)
     {
@@ -791,11 +931,19 @@ gtk_status_icon_update_image (GtkStatusIcon *status_icon)
 				 gtk_status_icon_blank_icon (status_icon));
 #endif
 #ifdef GDK_WINDOWING_WIN32
+      prev_hicon = priv->nid.hIcon;
       priv->nid.hIcon = gdk_win32_pixbuf_to_hicon_libgtk_only (gtk_status_icon_blank_icon (status_icon));
       priv->nid.uFlags |= NIF_ICON;
       if (priv->nid.hWnd != NULL && priv->visible)
 	if (!Shell_NotifyIconW (NIM_MODIFY, &priv->nid))
 	  g_warning ("%s:%d:Shell_NotifyIcon(NIM_MODIFY) failed", __FILE__, __LINE__-1);
+      if (prev_hicon)
+	DestroyIcon (prev_hicon);
+#endif
+#ifdef GDK_WINDOWING_QUARTZ
+      QUARTZ_POOL_ALLOC;
+      [priv->status_item setImage:gtk_status_icon_blank_icon (status_icon)];
+      QUARTZ_POOL_RELEASE;
 #endif
       return;
     }
@@ -832,12 +980,21 @@ gtk_status_icon_update_image (GtkStatusIcon *status_icon)
 	    gtk_image_set_from_pixbuf (GTK_IMAGE (priv->image), scaled);
 #endif
 #ifdef GDK_WINDOWING_WIN32
+	    prev_hicon = priv->nid.hIcon;
 	    priv->nid.hIcon = gdk_win32_pixbuf_to_hicon_libgtk_only (scaled);
 	    priv->nid.uFlags |= NIF_ICON;
 	    if (priv->nid.hWnd != NULL && priv->visible)
 	      if (!Shell_NotifyIconW (NIM_MODIFY, &priv->nid))
 		  g_warning ("%s:%d:Shell_NotifyIcon(NIM_MODIFY) failed", __FILE__, __LINE__-1);
+	    if (prev_hicon)
+	      DestroyIcon (prev_hicon);
 #endif
+#ifdef GDK_WINDOWING_QUARTZ
+      QUARTZ_POOL_ALLOC;
+      [priv->status_item setImage:scaled];
+      QUARTZ_POOL_RELEASE;
+#endif
+			
 	    g_object_unref (scaled);
 	  }
 	else
@@ -850,6 +1007,9 @@ gtk_status_icon_update_image (GtkStatusIcon *status_icon)
 	    if (priv->nid.hWnd != NULL && priv->visible)
 	      if (!Shell_NotifyIconW (NIM_MODIFY, &priv->nid))
 		g_warning ("%s:%d:Shell_NotifyIcon(NIM_MODIFY) failed", __FILE__, __LINE__-1);
+#endif
+#ifdef GDK_WINDOWING_QUARTZ
+      [priv->status_item setImage:NULL];
 #endif
 	  }
       }
@@ -870,13 +1030,31 @@ gtk_status_icon_update_image (GtkStatusIcon *status_icon)
 				    priv->image_data.stock_id,
 				    GTK_ICON_SIZE_SMALL_TOOLBAR,
 				    NULL);
+
+	  prev_hicon = priv->nid.hIcon;
 	  priv->nid.hIcon = gdk_win32_pixbuf_to_hicon_libgtk_only (pixbuf);
 	  priv->nid.uFlags |= NIF_ICON;
 	  if (priv->nid.hWnd != NULL && priv->visible)
 	    if (!Shell_NotifyIconW (NIM_MODIFY, &priv->nid))
 	      g_warning ("%s:%d:Shell_NotifyIcon(NIM_MODIFY) failed", __FILE__, __LINE__-1);
+	  if (prev_hicon)
+	    DestroyIcon (prev_hicon);
 	  g_object_unref (pixbuf);
 	}
+#endif
+#ifdef GDK_WINDOWING_QUARTZ
+	{
+	  GdkPixbuf *pixbuf;
+
+	  pixbuf = gtk_widget_render_icon (priv->dummy_widget,
+					   priv->image_data.stock_id,
+					   GTK_ICON_SIZE_SMALL_TOOLBAR,
+					   NULL);
+	  QUARTZ_POOL_ALLOC;
+	  [priv->status_item setImage:pixbuf];
+	  QUARTZ_POOL_RELEASE;
+	  g_object_unref (pixbuf);
+	}	
 #endif
       }
       break;
@@ -897,14 +1075,33 @@ gtk_status_icon_update_image (GtkStatusIcon *status_icon)
 				      priv->size,
 				      0, NULL);
 	  
+	  prev_hicon = priv->nid.hIcon;
 	  priv->nid.hIcon = gdk_win32_pixbuf_to_hicon_libgtk_only (pixbuf);
 	  priv->nid.uFlags |= NIF_ICON;
 	  if (priv->nid.hWnd != NULL && priv->visible)
 	    if (!Shell_NotifyIconW (NIM_MODIFY, &priv->nid))
 	      g_warning ("%s:%d:Shell_NotifyIcon(NIM_MODIFY) failed", __FILE__, __LINE__-1);
+	  if (prev_hicon)
+	    DestroyIcon (prev_hicon);
 	  g_object_unref (pixbuf);
 	}
 #endif
+#ifdef GDK_WINDOWING_QUARTZ
+	{
+	  GdkPixbuf *pixbuf;
+
+	  pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
+					     priv->image_data.icon_name,
+					     priv->size,
+					     0, NULL);
+
+	  QUARTZ_POOL_ALLOC;
+	  [priv->status_item setImage:pixbuf];
+	  QUARTZ_POOL_RELEASE;
+	  g_object_unref (pixbuf);
+	}
+#endif
+	
       }
       break;
       
@@ -917,6 +1114,13 @@ gtk_status_icon_update_image (GtkStatusIcon *status_icon)
       if (priv->nid.hWnd != NULL && priv->visible)
 	if (!Shell_NotifyIconW (NIM_MODIFY, &priv->nid))
 	  g_warning ("%s:%d:Shell_NotifyIcon(NIM_MODIFY) failed", __FILE__, __LINE__-1);
+#endif
+#ifdef GDK_WINDOWING_QUARTZ
+        {
+          QUARTZ_POOL_ALLOC;
+          [priv->status_item setImage:NULL];
+          QUARTZ_POOL_RELEASE;
+        }
 #endif
       break;
     default:
@@ -956,6 +1160,34 @@ gtk_status_icon_size_allocate (GtkStatusIcon *status_icon,
     }
 }
 
+static void
+gtk_status_icon_screen_changed (GtkStatusIcon *status_icon,
+				GdkScreen *old_screen)
+{
+  GtkStatusIconPrivate *priv = status_icon->priv;
+
+  if (gtk_widget_get_screen (priv->tray_icon) != old_screen)
+    {
+      g_object_notify (G_OBJECT (status_icon), "screen");
+    }
+}
+
+#endif
+
+#ifdef GDK_WINDOWING_X11
+
+static void
+gtk_status_icon_embedded_changed (GtkStatusIcon *status_icon)
+{
+  g_object_notify (G_OBJECT (status_icon), "embedded");
+}
+
+static void
+gtk_status_icon_orientation_changed (GtkStatusIcon *status_icon)
+{
+  g_object_notify (G_OBJECT (status_icon), "orientation");
+}
+
 #endif
 
 static gboolean
@@ -981,9 +1213,6 @@ gtk_status_icon_reset_image_data (GtkStatusIcon *status_icon)
 {
   GtkStatusIconPrivate *priv = status_icon->priv;
 
-  priv->storage_type = GTK_IMAGE_EMPTY;
-  g_object_notify (G_OBJECT (status_icon), "storage-type");
-
   switch (priv->storage_type)
   {
     case GTK_IMAGE_PIXBUF:
@@ -1013,6 +1242,9 @@ gtk_status_icon_reset_image_data (GtkStatusIcon *status_icon)
       g_assert_not_reached ();
       break;
   }
+
+  priv->storage_type = GTK_IMAGE_EMPTY;
+  g_object_notify (G_OBJECT (status_icon), "storage-type");
 }
 
 static void
@@ -1044,7 +1276,7 @@ gtk_status_icon_set_image (GtkStatusIcon *status_icon,
       g_object_notify (G_OBJECT (status_icon), "icon-name");
       break;
     default:
-      g_warning ("Image type %d not handled by GtkStatusIcon", storage_type);
+      g_warning ("Image type %u not handled by GtkStatusIcon", storage_type);
     }
 
   g_object_thaw_notify (G_OBJECT (status_icon));
@@ -1271,6 +1503,9 @@ gtk_status_icon_get_icon_name (GtkStatusIcon *status_icon)
  * if the size of the notification area changes. For other
  * storage types, the size-changed signal can be used to
  * react to size changes.
+ *
+ * Note that the returned size is only meaningful while the 
+ * status icon is embedded (see gtk_status_icon_is_embedded()).
  * 
  * Return value: the size that is available for the image
  *
@@ -1284,6 +1519,50 @@ gtk_status_icon_get_size (GtkStatusIcon *status_icon)
   return status_icon->priv->size;
 }
 
+/**
+ * gtk_status_icon_set_screen:
+ * @status_icon: a #GtkStatusIcon
+ * @screen: a #GdkScreen
+ *
+ * Sets the #GdkScreen where @status_icon is displayed; if
+ * the icon is already mapped, it will be unmapped, and
+ * then remapped on the new screen.
+ *
+ * Since: 2.12
+ */
+void
+gtk_status_icon_set_screen (GtkStatusIcon *status_icon,
+                            GdkScreen     *screen)
+{
+  g_return_if_fail (GDK_IS_SCREEN (screen));
+
+#ifdef GDK_WINDOWING_X11
+  gtk_window_set_screen (GTK_WINDOW (status_icon->priv->tray_icon), screen);
+#endif
+}
+
+/** 
+ * gtk_status_icon_get_screen:
+ * @window: a #GtkStatusIcon.
+ *
+ * Returns the #GdkScreen associated with @status_icon.
+ *
+ * Return value: a #GdkScreen.
+ *
+ * Since: 2.12
+ */
+GdkScreen *
+gtk_status_icon_get_screen (GtkStatusIcon *status_icon)
+{
+  g_return_val_if_fail (GTK_IS_STATUS_ICON (status_icon), NULL);
+
+#ifdef GDK_WINDOWING_X11   
+  return gtk_window_get_screen (GTK_WINDOW (status_icon->priv->tray_icon));
+#else
+  return gdk_screen_get_default ();
+#endif
+}
+
 /**
  * gtk_status_icon_set_tooltip:
  * @status_icon: a #GtkStatusIcon
@@ -1323,6 +1602,11 @@ gtk_status_icon_set_tooltip (GtkStatusIcon *status_icon,
     if (!Shell_NotifyIconW (NIM_MODIFY, &priv->nid))
       g_warning ("%s:%d:Shell_NotifyIconW(NIM_MODIFY) failed", __FILE__, __LINE__-1);
 #endif
+#ifdef GDK_WINDOWING_QUARTZ
+  QUARTZ_POOL_ALLOC;
+  [priv->status_item setToolTip:tooltip_text];
+  QUARTZ_POOL_RELEASE;
+#endif
 }
 
 static gboolean
@@ -1347,7 +1631,7 @@ gtk_status_icon_enable_blinking (GtkStatusIcon *status_icon)
       gtk_status_icon_blinker (status_icon);
 
       priv->blinking_timeout =
-	g_timeout_add (BLINK_TIMEOUT, 
+	gdk_threads_add_timeout (BLINK_TIMEOUT, 
 		       (GSourceFunc) gtk_status_icon_blinker, 
 		       status_icon);
     }
@@ -1396,8 +1680,10 @@ gtk_status_icon_set_visible (GtkStatusIcon *status_icon,
 #ifdef GDK_WINDOWING_X11
       if (visible)
 	gtk_widget_show (priv->tray_icon);
-      else
+      else if (GTK_WIDGET_REALIZED (priv->tray_icon)) {
 	gtk_widget_hide (priv->tray_icon);
+	gtk_widget_unrealize (priv->tray_icon);
+      }
 #endif
 #ifdef GDK_WINDOWING_WIN32
       if (priv->nid.hWnd != NULL)
@@ -1407,6 +1693,11 @@ gtk_status_icon_set_visible (GtkStatusIcon *status_icon,
 	  else
 	    Shell_NotifyIconW (NIM_DELETE, &priv->nid);
 	}
+#endif
+#ifdef GDK_WINDOWING_QUARTZ
+      QUARTZ_POOL_ALLOC;
+      [priv->status_item setVisible:visible];
+      QUARTZ_POOL_RELEASE;
 #endif
       g_object_notify (G_OBJECT (status_icon), "visible");
     }
@@ -1521,6 +1812,9 @@ gtk_status_icon_is_embedded (GtkStatusIcon *status_icon)
 #ifdef GDK_WINDOWING_WIN32
   return TRUE;
 #endif
+#ifdef GDK_WINDOWING_QUARTZ
+  return TRUE;
+#endif
 }
 
 /**
@@ -1629,6 +1923,21 @@ gtk_status_icon_position_menu (GtkMenu  *menu,
 
   *push_in = FALSE;
 #endif /* GDK_WINDOWING_X11 */
+
+#ifdef GDK_WINDOWING_WIN32
+  GtkStatusIcon *status_icon;
+  GtkStatusIconPrivate *priv;
+  
+  g_return_if_fail (GTK_IS_MENU (menu));
+  g_return_if_fail (GTK_IS_STATUS_ICON (user_data));
+
+  status_icon = GTK_STATUS_ICON (user_data);
+  priv = status_icon->priv;
+
+  *x = priv->last_click_x;
+  *y = priv->last_click_y;
+  *push_in = TRUE;
+#endif
 }
 
 /**
