@@ -58,6 +58,13 @@ enum
   DRAG_COLUMN_WINDOW_STATE_ARROW_RIGHT = 4
 };
 
+enum
+{
+  RUBBER_BAND_OFF = 0,
+  RUBBER_BAND_MAYBE_START = 1,
+  RUBBER_BAND_ACTIVE = 2
+};
+
 #define GTK_TREE_VIEW_SET_FLAG(tree_view, flag)   G_STMT_START{ (tree_view->priv->flags|=flag); }G_STMT_END
 #define GTK_TREE_VIEW_UNSET_FLAG(tree_view, flag) G_STMT_START{ (tree_view->priv->flags&=~(flag)); }G_STMT_END
 #define GTK_TREE_VIEW_FLAG_SET(tree_view, flag)   ((tree_view->priv->flags&flag)==flag)
@@ -114,6 +121,7 @@ struct _GtkTreeViewPrivate
   /* we cache it for simplicity of the code */
   gint dy;
   gint drag_column_x;
+  gint cursor_offset;
 
   GtkTreeViewColumn *expander_column;
   GtkTreeViewColumn *edited_column;
@@ -172,6 +180,18 @@ struct _GtkTreeViewPrivate
   gint press_start_x;
   gint press_start_y;
 
+  gint rubber_band_status;
+  gint rubber_band_x;
+  gint rubber_band_y;
+  gint rubber_band_shift;
+  gint rubber_band_ctrl;
+
+  GtkRBNode *rubber_band_start_node;
+  GtkRBTree *rubber_band_start_tree;
+
+  GtkRBNode *rubber_band_end_node;
+  GtkRBTree *rubber_band_end_tree;
+
   /* fixed height */
   gint fixed_height;
 
@@ -210,6 +230,10 @@ struct _GtkTreeViewPrivate
   guint hover_expand : 1;
   guint imcontext_changed : 1;
 
+  guint rubber_banding_enable : 1;
+
+  guint in_grab : 1;
+
 
   /* Auto expand/collapse timeout in hover mode */
   guint auto_expand_timeout;
@@ -234,6 +258,12 @@ struct _GtkTreeViewPrivate
   GtkDestroyNotify row_separator_destroy;
 
   gint level_indentation;
+
+  GtkTreeViewGridLines grid_lines;
+  GdkGC *grid_line_gc;
+
+  gboolean tree_lines_enabled;
+  GdkGC *tree_line_gc;
 };
 
 #ifdef __GNUC__
@@ -379,6 +409,10 @@ void		  _gtk_tree_view_column_cell_render      (GtkTreeViewColumn *tree_column,
 							  GdkRectangle      *cell_area,
 							  GdkRectangle      *expose_area,
 							  guint              flags);
+void		  _gtk_tree_view_column_get_focus_area   (GtkTreeViewColumn *tree_column,
+							  GdkRectangle      *background_area,
+							  GdkRectangle      *cell_area,
+							  GdkRectangle      *focus_area);
 gboolean	  _gtk_tree_view_column_cell_focus       (GtkTreeViewColumn *tree_column,
 							  gint               direction,
 							  gboolean           left,
